#ifndef DUNE_NPDE_P12D_HH
#define DUNE_NPDE_P12D_HH

#include<vector>
#include<iostream>
#include<dune/common/fvector.hh>
#include<dune/common/fmatrix.hh>
#include<dune/common/geometrytype.hh>
#include<dune/common/exceptions.hh>

#include<dune/grid/common/genericreferenceelements.hh>
#include<dune/grid/common/quadraturerules.hh>

#include<dune/localfunctions/common/localbasis.hh>
#include<dune/localfunctions/common/localkey.hh>
#include<dune/localfunctions/common/localfiniteelementtraits.hh>

#include<dune/pdelab/finiteelementmap/finiteelementmap.hh>

/** \brief Define the P1 Lagrange basis functions in 2d on the reference triangle
 *
 *  \tparam D Type to represent domain.
 *  \tparam R Type to represent range.
 */
template<class D, class R>
class P12dLocalFiniteElement
{
  //! \brief Class for the basis functions
  class P12dLocalBasis
  {
    typedef Dune::FieldVector<D,2> Coordinate;
    static const int dim = 2;
    Dune::GeometryType gt; // store geometry type for the basis
    std::size_t n;         // the number of basis functions
    std::vector<Coordinate> s;      // Lagrange points on the reference interval
  public:
    typedef Dune::LocalBasisTraits<D,dim,Coordinate,R,1,Dune::FieldVector<R,1>,Dune::FieldMatrix<R,1,dim>, 1> 
    Traits;

    //! \brief make a basis object for given polynomial degree
    P12dLocalBasis () : gt(Dune::GeometryType::simplex,dim), n(dim+1)
    {}

    //! \brief return number of basis functions
    std::size_t size () const { return n; }
    
    //! \brief Evaluate all shape functions at a given point in local coordinates
    inline void evaluateFunction (const typename Traits::DomainType& in,
				  std::vector<typename Traits::RangeType>& out) const { 
      out.resize(n);
      
      // lagrange point (0,0)
      out[0] = 1 - in[0] - in[1];

      // lagrange point (1,0)
      out[1] = in[0];

      // lagrange point (0,1)
      out[2] = in[1];
    }
    
    //! \brief Evaluate Jacobian of all shape functions
    inline void 
    evaluateJacobian (const typename Traits::DomainType& in,
		      std::vector<typename Traits::JacobianType>& out) const {  
      out.resize(n);

      // lagrange point (0,0)
      out[0][0][0] = -1.0;
      out[0][0][1] = -1.0;

      // lagrange point (1,0)
      out[1][0][0] = 1.0;
      out[1][0][1] = 0.0;

      // lagrange point (0,1)
      out[2][0][0] = 0.0;
      out[2][0][1] = 1.0;
    }
    
    //! \brief Polynomial order of the basis functions
    unsigned int order () const  {
      return 1;
    }

    //! \brief return geometry type
    Dune::GeometryType type () const { return gt; }
  };

  //! \brief Class for the basis functions 

  //! Enumeration of sub entities of an entity should correspond to
  //! http://www.dune-project.org/doc-2.0/doxygen/html/group__GridGenericReferenceElements.html
  //! Information about the meaning and syntax of the LocalKey class can be found on
  //! http://www.dune-project.org/doc/doxygen/dune-localfunctions-html/class_dune_1_1_local_key.html
  class P12dLocalCoefficients 
  {
    static const int dim = 2;
  public:
    P12dLocalCoefficients () : n(dim+1), li(n)  {

      // lagrange point (0,0)
      li[0] = Dune::LocalKey(0,dim,0);

      // lagrange point (1,0)
      li[1] = Dune::LocalKey(1,dim,0);

      // lagrange point (0,1)
      li[2] = Dune::LocalKey(2,dim,0);
    }
    
    //! number of coefficients
    std::size_t size () const { return n; }
    
    //! map index i to local key
    const Dune::LocalKey& localKey (int i) const  {
      return li[i];
    } 
    
  private:
    std::size_t n;                  // the number of basis functions
    std::vector<Dune::LocalKey> li; // assignment of basis function to subentities
  };

  //! \brief Class for interpolating a given function by the basis
  template<typename LB>
  class P12dLocalInterpolation
  {
    static const int dim = 2;
  public:
    P12dLocalInterpolation () : n(dim+1) {}

    //! \brief Local interpolation of a function. Sets the degrees of
    //! freedom according to the interpolation of a given function f
    template<typename F, typename C>
    void interpolate (const F& f, std::vector<C>& out) const 
    {
      out.resize(n);
      typename LB::Traits::DomainType x;
      typename LB::Traits::RangeType y;

      // lagrange point (0,0)
      x = 0;
      f.evaluate(x,y);
      out[0] = y;

      // lagrange point (1,0)
      x = 0; x[0] = 1.0;
      f.evaluate(x,y);
      out[1] = y;

      // lagrange point (0,1)
      x = 0; x[1] = 1.0;
      f.evaluate(x,y);
      out[2] = y;

    }
  private:
    std::size_t n;                  // the number of basis functions
  };

  Dune::GeometryType gt;
  P12dLocalBasis basis;
  P12dLocalCoefficients coefficients;
  P12dLocalInterpolation<P12dLocalBasis> interpolation;

public:
  typedef Dune::LocalFiniteElementTraits<P12dLocalBasis,
					 P12dLocalCoefficients,
					 P12dLocalInterpolation<P12dLocalBasis> > Traits;

  P12dLocalFiniteElement () 
    : gt(Dune::GeometryType::simplex,2)
  {}

  const typename Traits::LocalBasisType& localBasis () const 
  {
    return basis;
  }
    
  const typename Traits::LocalCoefficientsType& localCoefficients () const 
  {
    return coefficients;
  }
    
  const typename Traits::LocalInterpolationType& localInterpolation () const 
  {
    return interpolation;
  }
    
  Dune::GeometryType type () const { return gt; }
    
  P12dLocalFiniteElement* clone () const {
    return new P12dLocalFiniteElement(*this);
  }
};

/** \brief FiniteElementMap for the P1 basis in 2d
 *
 *  \tparam D Type to represent domain.
 *  \tparam R Type to represent range.
 */
template<class D, class R>
class P12dLocalFiniteElementMap
  : public Dune::PDELab::SimpleLocalFiniteElementMap< P12dLocalFiniteElement<D,R> >
{
public:
  P12dLocalFiniteElementMap ()
    : Dune::PDELab::SimpleLocalFiniteElementMap< P12dLocalFiniteElement<D,R> >(P12dLocalFiniteElement<D,R>())
  {}
};

#endif
