// -*- tab-width: 4; indent-tabs-mode: nil -*-
#ifndef DUNE_NPDE_PK1DCONSTRAINTS_HH
#define DUNE_NPDE_PK1DCONSTRAINTS_HH

#include <cstddef>
#include <dune/common/exceptions.hh>
#include <dune/grid/common/genericreferenceelements.hh>
#include <dune/grid/common/grid.hh>
#include <dune/common/geometrytype.hh>
#include <dune/pdelab/common/geometrywrapper.hh>
#include <dune/pdelab/gridfunctionspace/gridfunctionspace.hh>
#include <dune/pdelab/gridfunctionspace/genericdatahandle.hh>

//! \brief Dirichlet constraints for Pk1d
class Pk1dConstraints
{
public:
  enum { doBoundary = true };
  enum { doProcessor = false };
  enum { doSkeleton = false };
  enum { doVolume = false };

  //! boundary constraints
  /**
   * \tparam F   Boundary grid function returning boundary condition type.
   *             If the value of the boundary grid function evaluated at
   *             the local face center is >0, a dirichlet boundary is
   *             assumed.
   * \tparam IG  intersection geometry
   * \tparam LFS local function space
   * \tparam T   TransformationType
   */
  template<typename F, typename I, typename LFS, typename T>
  void boundary (const F& f, const Dune::PDELab::IntersectionGeometry<I>& ig, 
		 const LFS& lfs, T& trafo) const
  {
    typename F::Traits::RangeType bctype;

    const int face = ig.indexInInside();

    // find all local indices of this face
    Dune::GeometryType gt = ig.inside()->type();
    typedef typename Dune::PDELab::IntersectionGeometry<I>::ctype DT;
    const int dim = Dune::PDELab::IntersectionGeometry<I>::Entity::Geometry::dimension;
    const Dune::GenericReferenceElement<DT,dim>& refelem = Dune::GenericReferenceElements<DT,dim>::general(gt);

    const Dune::GenericReferenceElement<DT,dim-1> & 
      face_refelem = Dune::GenericReferenceElements<DT,dim-1>::general(ig.geometry().type()); 

    // empty map means Dirichlet constraint
    typename T::RowType empty;

    const typename F::Traits::DomainType testpoint = face_refelem.position(0,0);
    f.evaluate(ig,testpoint,bctype);
    if (bctype<=0) return;

    // find basis function that is associated with this corner
    for (std::size_t i=0; i<lfs.localFiniteElement().localCoefficients().size(); i++)
      {
	// The codim to which this dof is attached to
	unsigned int codim = lfs.localFiniteElement().localCoefficients().localKey(i).codim();
	if (codim==0) continue;

	if ((int) lfs.localFiniteElement().localCoefficients().localKey(i).subEntity()==face)
	  trafo[i] = empty;
      }
  }
};

#endif
