#include <iostream>
#include "config.h"           // file constructed by ./configure script
#include <dune/common/mpihelper.hh> // include mpi helper class 
#include <dune/common/configparser.hh>  
#include <dune/grid/sgrid.hh> // load sgrid definition
#include <dune/grid/onedgrid.hh>
#include <dune/grid/io/file/vtk/subsamplingvtkwriter.hh>
#if HAVE_UG 
#include <dune/grid/uggrid.hh>
#endif
#include<dune/grid/io/file/gmshreader.hh>
#include<dune/istl/bvector.hh>
#include<dune/istl/operators.hh>
#include<dune/istl/solvers.hh>
#include<dune/istl/preconditioners.hh>
#include<dune/istl/io.hh>
#include<dune/istl/paamg/amg.hh>
#include <dune/pdelab/common/function.hh>
#include <dune/pdelab/common/vtkexport.hh>
#include <dune/pdelab/backend/istlvectorbackend.hh>
#include <dune/pdelab/backend/istlmatrixbackend.hh>
#include <dune/pdelab/backend/istlsolverbackend.hh>
#include <dune/pdelab/gridfunctionspace/gridfunctionspace.hh>
#include <dune/pdelab/gridfunctionspace/gridfunctionspaceutilities.hh>
#include <dune/pdelab/gridfunctionspace/interpolate.hh>
#include <dune/pdelab/gridfunctionspace/constraints.hh>
#include <dune/pdelab/gridoperatorspace/gridoperatorspace.hh>
#include <dune/pdelab/stationary/linearproblem.hh>
#include<dune/pdelab/finiteelementmap/pk2dfem.hh>
#include<dune/pdelab/finiteelementmap/conformingconstraints.hh>

#include "../../dune/npde/finiteelement/pk1dbasis.hh"
#include "../../dune/npde/finiteelement/pk1dconstraints.hh"
#include "../../dune/npde/localoperator/convectiondiffusionfem.hh"

template<typename GV, typename RF>
class ConvectionDiffusion_Example113_Wathenetal
{
  typedef ConvectionDiffusionBoundaryConditions::Type BCType;

public:
  typedef ConvectionDiffusionParameterTraits<GV,RF> Traits;

  //! tensor diffusion coefficient
  typename Traits::PermTensorType
  A (const typename Traits::ElementType& e, const typename Traits::DomainType& x) const
  {
    typename Traits::PermTensorType I;
    for (std::size_t i=0; i<Traits::dimDomain; i++)
      for (std::size_t j=0; j<Traits::dimDomain; j++)
        I[i][j] = (i==j) ? 1 : 0;
    return I;
  }

  //! velocity field
  typename Traits::RangeType
  b (const typename Traits::ElementType& e, const typename Traits::DomainType& x) const
  {
    typename Traits::RangeType v(0.0);
    return v;
  }

  //! sink term
  typename Traits::RangeFieldType 
  c (const typename Traits::ElementType& e, const typename Traits::DomainType& x) const
  {
    return 0.0;
  }

  //! source term
  typename Traits::RangeFieldType 
  f (const typename Traits::ElementType& e, const typename Traits::DomainType& x) const
  {
    return 0.0;
  }

  //! boundary condition type function
  BCType
  bctype (const typename Traits::IntersectionType& is, const typename Traits::IntersectionDomainType& x) const
  {
    typename Traits::DomainType xglobal = is.geometry().global(x);
    return ConvectionDiffusionBoundaryConditions::Dirichlet;
  }

  //! Dirichlet boundary condition value
  typename Traits::RangeFieldType 
  g (const typename Traits::ElementType& e, const typename Traits::DomainType& xlocal) const
  {
    typename Traits::DomainType x = e.geometry().global(xlocal);
    return 2.0*(1.0+x[1])/( (3.0+x[0])*(3.0+x[0]) + (1.0+x[1])*(1.0+x[1]));
  }

  //! Neumann boundary condition
  typename Traits::RangeFieldType 
  j (const typename Traits::IntersectionType& is, const typename Traits::IntersectionDomainType& x) const
  {
    return 0.0;
  }

  //! Neumann boundary condition
  typename Traits::RangeFieldType 
  o (const typename Traits::IntersectionType& is, const typename Traits::IntersectionDomainType& x) const
  {
    return 0.0;
  }
};

template<typename GV, typename RF>
class ReentrantCornerProblem
{
  typedef ConvectionDiffusionBoundaryConditions::Type BCType;

public:
  typedef ConvectionDiffusionParameterTraits<GV,RF> Traits;

  //! tensor diffusion coefficient
  typename Traits::PermTensorType
  A (const typename Traits::ElementType& e, const typename Traits::DomainType& x) const
  {
    typename Traits::PermTensorType I;
    for (std::size_t i=0; i<Traits::dimDomain; i++)
      for (std::size_t j=0; j<Traits::dimDomain; j++)
        I[i][j] = (i==j) ? 1 : 0;
    return I;
  }

  //! velocity field
  typename Traits::RangeType
  b (const typename Traits::ElementType& e, const typename Traits::DomainType& x) const
  {
    typename Traits::RangeType v(0.0);
    return v;
  }

  //! sink term
  typename Traits::RangeFieldType 
  c (const typename Traits::ElementType& e, const typename Traits::DomainType& x) const
  {
    return 0.0;
  }

  //! source term
  typename Traits::RangeFieldType 
  f (const typename Traits::ElementType& e, const typename Traits::DomainType& x) const
  {
    return 0.0;
  }

  //! boundary condition type function
  BCType
  bctype (const typename Traits::IntersectionType& is, const typename Traits::IntersectionDomainType& x) const
  {
    typename Traits::DomainType xglobal = is.geometry().global(x);
    return ConvectionDiffusionBoundaryConditions::Dirichlet;
  }

  //! Dirichlet boundary condition value
  typename Traits::RangeFieldType 
  g (const typename Traits::ElementType& e, const typename Traits::DomainType& xlocal) const
  {
    typename Traits::DomainType x = e.geometry().global(xlocal);

    typename Traits::DomainFieldType theta = std::atan2(x[1], x[0]);
    if(theta < 0.0) theta += 2*M_PI;
    typename Traits::DomainFieldType r = x.two_norm();

    if (theta<1e-5 || theta>1.5*M_PI-1e-5)
      return 0.0;
    else
      return pow(r,2.0/3.0)*std::sin(theta*2.0/3.0);
  }

  //! Neumann boundary condition
  typename Traits::RangeFieldType 
  j (const typename Traits::IntersectionType& is, const typename Traits::IntersectionDomainType& x) const
  {
    return 0.0;
  }

  //! Neumann boundary condition
  typename Traits::RangeFieldType 
  o (const typename Traits::IntersectionType& is, const typename Traits::IntersectionDomainType& x) const
  {
    return 0.0;
  }
};

//! solve the diffusion problem on a given GridView with a given finite element space
template<class GV, class FEM, class CON, class Problem>
void driver (const GV& gv, FEM& fem, Problem& problem, std::string filename_base)
{
  // coordinate and result type
  typedef typename GV::Grid::ctype Coord;
  typedef double Real;

  // make grid function space 
  typedef Dune::PDELab::ISTLVectorBackend<1> VBE;
  typedef Dune::PDELab::GridFunctionSpace<GV,FEM,CON,VBE> GFS;
  GFS gfs(gv,fem);

  // make a std::vector of degree of freedom vectors and initialize it with Dirichlet extension
  typedef typename GFS::template VectorContainer<Real>::Type U;
  U u(gfs,0.0);
  ConvectionDiffusionDirichletExtensionAdapter<Problem> g(gv,problem);
  Dune::PDELab::interpolate(g,gfs,u);

  // make constraints container and initialize it
  typedef typename GFS::template ConstraintsContainer<Real>::Type CC;
  CC cc;
  ConvectionDiffusionBoundaryConditionAdapter<Problem> bctype(gv,problem);
  Dune::PDELab::constraints(bctype,gfs,cc);
  Dune::PDELab::set_nonconstrained_dofs(cc,0.0,u);
  std::cout << "constrained dofs=" << cc.size() << " of " << gfs.globalSize() << std::endl;

  // make local operator
  typedef ConvectionDiffusionFEM<Problem> LOP;
  LOP lop(problem);
  typedef Dune::PDELab::ISTLBCRSMatrixBackend<1,1> MBE;
  typedef Dune::PDELab::GridOperatorSpace<GFS,GFS,LOP,CC,CC,MBE> GOS;
  GOS gos(gfs,cc,gfs,cc,lop);

  // make linear solver and solve problem
  typedef Dune::PDELab::ISTLBackend_SEQ_SuperLU LS;
  LS ls(true);
  typedef Dune::PDELab::StationaryLinearProblemSolver<GOS,LS,U> SLP;
  SLP slp(gos,u,ls,1e-6);
  slp.apply();

  // write vtk file
  typedef Dune::PDELab::DiscreteGridFunction<GFS,U> DGF;
  DGF udgf(gfs,u);
  //Dune::SubsamplingVTKWriter<GV> vtkwriter(gv,3);
  Dune::VTKWriter<GV> vtkwriter(gv);
  vtkwriter.addVertexData(new Dune::PDELab::VTKGridFunctionAdapter<DGF>(udgf,"u"));
  vtkwriter.write(filename_base,Dune::VTKOptions::binaryappended);
}

int main(int argc, char **argv)
{
  // initialize MPI, finalize is done automatically on exit 
  Dune::MPIHelper::instance(argc,argv);

  // start try/catch block to get error messages from dune
  try {

    Dune::ConfigParser configuration;
    configuration.parseFile("diffusionproblem.ini");

    // OneDGrid/Pk1d version
    if (false)
    {
      // The continuous floating point type of the grid.
      typedef Dune::OneDGrid::ctype ctype;
      
      // Set number of vertices
      const unsigned int N = configuration.get<int>("grid.cells");
      const ctype L = 1.0;
      
      // Define position intervals in reference domain [0,1].
      typedef std::vector<ctype> Intervals;
      Intervals intervals(N+1);
      for(unsigned int i=0; i<N+1; ++i)
	intervals[i] = ctype(i) / ctype(N);
      
      // Construct grid
      Dune::OneDGrid grid(intervals);
      
      // call generic function
      typedef Dune::OneDGrid::LeafGridView GV;
      const GV& gv=grid.leafView();

      // make the finite element map
      typedef Pk1dLocalFiniteElementMap<Dune::OneDGrid::ctype,double> FEM;
      FEM fem(2);
      typedef Pk1dConstraints CON;
      typedef ConvectionDiffusionModelProblem<GV,double> Problem;
      Problem problem;
      std::string filename_base = "diffusion_onedgrid";
      driver<GV,FEM,CON,Problem>(gv,fem,problem,filename_base);
    }

    // UGGrid/Pk2d version
#if HAVE_UG
    if (true)
    {
      // make uggrid
      typedef Dune::UGGrid<2> GridType;
      GridType grid;
      typedef std::vector<int> GmshIndexMap;
      GmshIndexMap boundary_index_map;
      GmshIndexMap element_index_map;
      std::string grid_file = configuration.get<std::string>("grid.filename");
      Dune::GmshReader<GridType> gmsh_reader;
      gmsh_reader.read(grid,grid_file,boundary_index_map,element_index_map,true,false);
      const int max_level = configuration.get<int>("grid.max_level");
      for (int i=0; i<max_level; i++) grid.globalRefine(1);
      typedef GridType::LeafGridView GV;
      const GV& gv=grid.leafView();

      // make finite element space and solve
      typedef Dune::PDELab::Pk2DLocalFiniteElementMap<GV,GridType::ctype,double,1> FEM;
      FEM fem(gv);
      typedef Dune::PDELab::ConformingDirichletConstraints CON;
      typedef ReentrantCornerProblem<GV,double> Problem;
      Problem problem;
      const std::string filename_base = "diffusion_uggrid2d";
      driver<GV,FEM,CON,Problem>(gv,fem,problem,filename_base);
    }
#endif
  }
  catch (std::exception & e) {
    std::cout << "STL ERROR: " << e.what() << std::endl;
    return 1;
  }
  catch (Dune::Exception & e) {
    std::cout << "DUNE ERROR: " << e.what() << std::endl;
    return 1;
  }
  catch (...) {
    std::cout << "Unknown ERROR" << std::endl;
    return 1;
  }

  // done
  return 0;
}
