#include <iostream>
#include "config.h"           // file constructed by ./configure script
#include <dune/common/mpihelper.hh> // include mpi helper class 
#include <dune/common/configparser.hh>  
#include <dune/grid/sgrid.hh> // load sgrid definition
#include <dune/grid/onedgrid.hh>
#include <dune/grid/io/file/vtk/subsamplingvtkwriter.hh>
#include<dune/grid/yaspgrid.hh>
#if HAVE_UG 
#include <dune/grid/uggrid.hh>
#endif
#include<dune/grid/io/file/gmshreader.hh>
#include<dune/common/geometrytype.hh>
#include<dune/grid/common/quadraturerules.hh>
#include<dune/istl/bvector.hh>
#include<dune/istl/operators.hh>
#include<dune/istl/solvers.hh>
#include<dune/istl/preconditioners.hh>
#include<dune/istl/io.hh>
#include<dune/istl/paamg/amg.hh>
#include <dune/pdelab/common/function.hh>
#include <dune/pdelab/common/functionutilities.hh>
#include <dune/pdelab/common/vtkexport.hh>
#include <dune/pdelab/backend/istlvectorbackend.hh>
#include <dune/pdelab/backend/istlmatrixbackend.hh>
#include <dune/pdelab/backend/istlsolverbackend.hh>
#include <dune/pdelab/gridfunctionspace/gridfunctionspace.hh>
#include <dune/pdelab/gridfunctionspace/gridfunctionspaceutilities.hh>
#include <dune/pdelab/gridfunctionspace/interpolate.hh>
#include <dune/pdelab/gridfunctionspace/constraints.hh>
#include <dune/pdelab/gridoperatorspace/gridoperatorspace.hh>
#include <dune/pdelab/stationary/linearproblem.hh>
#include<dune/pdelab/finiteelementmap/pk2dfem.hh>
#include<dune/pdelab/finiteelementmap/q22dfem.hh>
#include<dune/pdelab/finiteelementmap/q1fem.hh>
#include<dune/pdelab/finiteelementmap/conformingconstraints.hh>

#include "../../dune/npde/finiteelement/pk1dbasis.hh"
#include "../../dune/npde/finiteelement/pk1dconstraints.hh"
#include "../../dune/npde/localoperator/convectiondiffusionfem.hh"
#include "../../dune/npde/utilities/functionutilities.hh"
#include "../../dune/npde/utilities/meshsize.hh"

template<typename GV, typename RF>
class ConvectionDiffusion_Example113_Wathenetal
{
  typedef ConvectionDiffusionBoundaryConditions::Type BCType;

public:
  typedef ConvectionDiffusionParameterTraits<GV,RF> Traits;

  //! tensor diffusion coefficient
  typename Traits::PermTensorType
  A (const typename Traits::ElementType& e, const typename Traits::DomainType& x) const
  {
    typename Traits::PermTensorType I;
    for (std::size_t i=0; i<Traits::dimDomain; i++)
      for (std::size_t j=0; j<Traits::dimDomain; j++)
        I[i][j] = (i==j) ? 1 : 0;
    return I;
  }

  //! velocity field
  typename Traits::RangeType
  b (const typename Traits::ElementType& e, const typename Traits::DomainType& x) const
  {
    typename Traits::RangeType v(0.0);
    return v;
  }

  //! sink term
  typename Traits::RangeFieldType 
  c (const typename Traits::ElementType& e, const typename Traits::DomainType& x) const
  {
    return 0.0;
  }

  //! source term
  typename Traits::RangeFieldType 
  f (const typename Traits::ElementType& e, const typename Traits::DomainType& x) const
  {
    return 0.0;
  }

  //! boundary condition type function
  BCType
  bctype (const typename Traits::IntersectionType& is, const typename Traits::IntersectionDomainType& x) const
  {
    typename Traits::DomainType xglobal = is.geometry().global(x);
    return ConvectionDiffusionBoundaryConditions::Dirichlet;
  }

  //! Dirichlet boundary condition value
  typename Traits::RangeFieldType 
  g (const typename Traits::ElementType& e, const typename Traits::DomainType& xlocal) const
  {
    typename Traits::DomainType x = e.geometry().global(xlocal);
    return 2.0*(x[1])/( (2.0+x[0])*(2.0+x[0]) + x[1]*x[1] );
  }

  //! Neumann boundary condition
  typename Traits::RangeFieldType 
  j (const typename Traits::IntersectionType& is, const typename Traits::IntersectionDomainType& x) const
  {
    return 0.0;
  }

  //! Neumann boundary condition
  typename Traits::RangeFieldType 
  o (const typename Traits::IntersectionType& is, const typename Traits::IntersectionDomainType& x) const
  {
    return 0.0;
  }
};


//! exact gradient of solution
template<typename GV, typename RF>
class ExactGradient
  : public Dune::PDELab::GridFunctionBase<Dune::PDELab::GridFunctionTraits<GV,RF,
      GV::dimension,Dune::FieldVector<RF,GV::dimension> >,
      ExactGradient<GV,RF> >
{
  const GV& gv;

public:
  typedef Dune::PDELab::GridFunctionTraits<GV,RF,
      GV::dimension,Dune::FieldVector<RF,GV::dimension> > Traits;
  typedef Dune::PDELab::GridFunctionBase<Traits,ExactGradient<GV,RF> > BaseT;

  ExactGradient (const GV& gv_) : gv(gv_) {}

  inline void evaluate (const typename Traits::ElementType& e, 
                        const typename Traits::DomainType& xlocal,
                        typename Traits::RangeType& y) const
  {  
    typename Traits::DomainType x = e.geometry().global(xlocal);
    RF denom = (2.0+x[0])*(2.0+x[0])+x[1]*x[1];
    y[0] = -4.0*x[1]*(2.0+x[0])/(denom*denom);
    y[1] = 2.0/denom - 4.0*x[1]*x[1]/(denom*denom);
  }
  
  inline const typename Traits::GridViewType& getGridView () const
  {
    return gv;
  }
};


//! solve the diffusion problem on a given GridView with a given finite element space
template<class GV, class FEM, class CON>
void driver (const GV& gv, FEM& fem, std::string filename_base, int& N, double& l2e, double& h1se)
{
  // coordinate and result type
  typedef typename GV::Grid::ctype Coord;
  typedef double Real;

  // make grid function space 
  typedef Dune::PDELab::ISTLVectorBackend<1> VBE;
  typedef Dune::PDELab::GridFunctionSpace<GV,FEM,CON,VBE> GFS;
  GFS gfs(gv,fem);
  N = gfs.globalSize();

  // make problem parameters
  typedef ConvectionDiffusion_Example113_Wathenetal<GV,double> Problem;
  Problem problem;

  // make a std::vector of degree of freedom vectors and initialize it with Dirichlet extension
  typedef typename GFS::template VectorContainer<Real>::Type U;
  U u(gfs,0.0);
  typedef ConvectionDiffusionDirichletExtensionAdapter<Problem> G;
  G g(gv,problem);
  Dune::PDELab::interpolate(g,gfs,u);

  // make constraints container and initialize it
  typedef typename GFS::template ConstraintsContainer<Real>::Type CC;
  CC cc;
  ConvectionDiffusionBoundaryConditionAdapter<Problem> bctype(gv,problem);
  Dune::PDELab::constraints(bctype,gfs,cc);
  Dune::PDELab::set_nonconstrained_dofs(cc,0.0,u);
  std::cout << "constrained dofs=" << cc.size() << " of " << gfs.globalSize() << std::endl;

  // make local operator
  typedef ConvectionDiffusionFEM<Problem> LOP;
  LOP lop(problem);
  typedef Dune::PDELab::ISTLBCRSMatrixBackend<1,1> MBE;
  typedef Dune::PDELab::GridOperatorSpace<GFS,GFS,LOP,CC,CC,MBE> GOS;
  GOS gos(gfs,cc,gfs,cc,lop);

  // make linear solver and solve problem
  typedef Dune::PDELab::ISTLBackend_SEQ_CG_SSOR LS;
  LS ls;
  typedef Dune::PDELab::StationaryLinearProblemSolver<GOS,LS,U> SLP;
  SLP slp(gos,u,ls,1e-6);
  slp.apply();

  // compute errors
  typedef Dune::PDELab::DiscreteGridFunction<GFS,U> DGF;
  DGF udgf(gfs,u);
  typedef DifferenceSquaredAdapter<G,DGF> DifferenceSquared;
  DifferenceSquared differencesquared(g,udgf);
  typename DifferenceSquared::Traits::RangeType l2errorsquared(0.0);
  Dune::PDELab::integrateGridFunction(differencesquared,l2errorsquared,10);
  l2e = sqrt(l2errorsquared);
  typedef Dune::PDELab::DiscreteGridFunctionGradient<GFS,U> DGFGrad;
  DGFGrad udgfgrad(gfs,u);
  typedef ExactGradient<GV,double> Grad;
  Grad grad(gv);
  typedef DifferenceSquaredAdapter<Grad,DGFGrad> GradDifferenceSquared;
  GradDifferenceSquared graddifferencesquared(grad,udgfgrad);
  typename GradDifferenceSquared::Traits::RangeType h1semierrorsquared(0.0);
  Dune::PDELab::integrateGridFunction(graddifferencesquared,h1semierrorsquared,10);
  h1se = sqrt(h1semierrorsquared);

  // write vtk file
  typedef DifferenceAdapter<G,DGF> Difference;
  Difference difference(g,udgf);
  //Dune::SubsamplingVTKWriter<GV> vtkwriter(gv,6);
  Dune::VTKWriter<GV> vtkwriter(gv);
  vtkwriter.addVertexData(new Dune::PDELab::VTKGridFunctionAdapter<DGF>(udgf,"u_h"));
  vtkwriter.addVertexData(new Dune::PDELab::VTKGridFunctionAdapter<G>(g,"u"));
  vtkwriter.addVertexData(new Dune::PDELab::VTKGridFunctionAdapter<Difference>(difference,"u-u_h"));
  vtkwriter.write(filename_base,Dune::VTKOptions::binaryappended);
}


int main(int argc, char **argv)
{
  // initialize MPI, finalize is done automatically on exit 
  Dune::MPIHelper::instance(argc,argv);

  // start try/catch block to get error messages from dune
  try {

    Dune::ConfigParser configuration;
    configuration.parseFile("square.ini");
    std::string grid_file = configuration.get<std::string>("grid.filename");
    const int max_level = configuration.get<int>("grid.max_level");
    const int degree = configuration.get<int>("method.degree");

    std::vector<double> l2(max_level+1,0.0);
    std::vector<double> h1s(max_level+1,0.0);
    std::vector<int> N(max_level+1,0);
    std::vector<double> h(max_level+1,0.0);

    // YaspGrid/Qk2d version
    if (grid_file=="yasp")
    {
      // make YaspGrid
      Dune::FieldVector<double,2> L(2.0);
      Dune::FieldVector<int,2> S(2);
      Dune::FieldVector<bool,2> B(false);
      typedef Dune::YaspGrid<2> GridType;
      GridType grid(L,S,B,0);

      for (int i=0; i<=max_level; i++) 
	{
	  typedef GridType::LeafGridView GV;
	  const GV& gv=grid.leafView();
	  h[i] = meshsize(gv);

	  // make finite element space and solve
	  typedef Dune::PDELab::ConformingDirichletConstraints CON;
	  if (degree==1) {
	    typedef Dune::PDELab::Q1LocalFiniteElementMap<GridType::ctype,double,2> FEM;
	    FEM fem;
	    std::stringstream filename;
	    filename << "square_yaspgrid" << "_l" << i << "_k" << degree;
	    driver<GV,FEM,CON>(gv,fem,filename.str(),N[i],l2[i],h1s[i]);
	  }
	  if (degree==2) {
	    typedef Dune::PDELab::Q22DLocalFiniteElementMap<GridType::ctype,double> FEM;
	    FEM fem;
	    std::stringstream filename;
	    filename << "square_yaspgrid" << "_l" << i << "_k" << degree;
	    driver<GV,FEM,CON>(gv,fem,filename.str(),N[i],l2[i],h1s[i]);
	  }
	  if (i<max_level) grid.globalRefine(1);
	}
    }

    // UGGrid/Pk2d version
#if HAVE_UG
    if (grid_file!="yasp")
    {
      // make uggrid
      typedef Dune::UGGrid<2> GridType;
      GridType grid;
      typedef std::vector<int> GmshIndexMap;
      GmshIndexMap boundary_index_map;
      GmshIndexMap element_index_map;
      Dune::GmshReader<GridType> gmsh_reader;
      gmsh_reader.read(grid,grid_file,boundary_index_map,element_index_map,true,false);

      for (int i=0; i<=max_level; i++) 
	{
	  typedef GridType::LeafGridView GV;
	  const GV& gv=grid.leafView();
	  h[i] = meshsize(gv);

	  // make finite element space and solve
	  typedef Dune::PDELab::ConformingDirichletConstraints CON;
	  if (degree==1) {
	    typedef Dune::PDELab::Pk2DLocalFiniteElementMap<GV,GridType::ctype,double,1> FEM;
	    FEM fem(gv);
	    std::stringstream filename;
	    filename << "square_" << grid_file << "_l" << i << "_k" << degree;
	    driver<GV,FEM,CON>(gv,fem,filename.str(),N[i],l2[i],h1s[i]);
	  }
	  if (degree==2) {
	    typedef Dune::PDELab::Pk2DLocalFiniteElementMap<GV,GridType::ctype,double,2> FEM;
	    FEM fem(gv);
	    std::stringstream filename;
	    filename << "square_" << grid_file << "_l" << i << "_k" << degree;
	    driver<GV,FEM,CON>(gv,fem,filename.str(),N[i],l2[i],h1s[i]);
	  }
	  if (degree==3) {
	    typedef Dune::PDELab::Pk2DLocalFiniteElementMap<GV,GridType::ctype,double,3> FEM;
	    FEM fem(gv);
	    std::stringstream filename;
	    filename << "square_" << grid_file << "_l" << i << "_k" << degree;
	    driver<GV,FEM,CON>(gv,fem,filename.str(),N[i],l2[i],h1s[i]);
	  }
	  if (degree==4) {
	    typedef Dune::PDELab::Pk2DLocalFiniteElementMap<GV,GridType::ctype,double,4> FEM;
	    FEM fem(gv);
	    std::stringstream filename;
	    filename << "square_" << grid_file << "_l" << i << "_k" << degree;
	    driver<GV,FEM,CON>(gv,fem,filename.str(),N[i],l2[i],h1s[i]);
	  }
	  if (degree==5) {
	    typedef Dune::PDELab::Pk2DLocalFiniteElementMap<GV,GridType::ctype,double,5> FEM;
	    FEM fem(gv);
	    std::stringstream filename;
	    filename << "square_" << grid_file << "_l" << i << "_k" << degree;
	    driver<GV,FEM,CON>(gv,fem,filename.str(),N[i],l2[i],h1s[i]);
	  }
	  if (i<max_level) grid.globalRefine(1);
	}
    }
#endif
    std::cout << "Results on mesh=" << grid_file << " k=" << degree << std::endl;
    std::cout << "N 1/h l2 l2rate h1semi h1semirate" << std::endl;
    for (int i=0; i<=max_level; i++) 
      {
	double rate1=0.0;
	if (i>0) rate1=log(l2[i]/l2[i-1])/log(0.5);
	double rate2=0.0;
	if (i>0) rate2=log(h1s[i]/h1s[i-1])/log(0.5); 
	std::cout << std::setw(12) << N[i]
		  << std::setw(12) << std::setprecision(4) << std::scientific << 1.0/h[i]
		  << std::setw(12) << std::setprecision(4) << std::scientific << l2[i]
		  << std::setw(12) << std::setprecision(4) << std::scientific << rate1
		  << std::setw(12) << std::setprecision(4) << std::scientific << h1s[i]
		  << std::setw(12) << std::setprecision(4) << std::scientific << rate2
		  << std::endl;
	}
  }
  catch (std::exception & e) {
    std::cout << "STL ERROR: " << e.what() << std::endl;
    return 1;
  }
  catch (Dune::Exception & e) {
    std::cout << "DUNE ERROR: " << e.what() << std::endl;
    return 1;
  }
  catch (...) {
    std::cout << "Unknown ERROR" << std::endl;
    return 1;
  }

  // done
  return 0;
}
