#include <iostream>
#include "config.h"           // file constructed by ./configure script
#include <dune/common/mpihelper.hh> // include mpi helper class 
#include <dune/common/configparser.hh>  
#include <dune/grid/sgrid.hh> // load sgrid definition
#include <dune/grid/onedgrid.hh>
#include <dune/grid/io/file/vtk/subsamplingvtkwriter.hh>
#include <dune/pdelab/common/function.hh>
#include <dune/pdelab/common/vtkexport.hh>
#include <dune/pdelab/backend/istlvectorbackend.hh>
#include <dune/pdelab/backend/istlmatrixbackend.hh>
#include <dune/pdelab/backend/istlsolverbackend.hh>
#include <dune/pdelab/gridfunctionspace/gridfunctionspace.hh>
#include <dune/pdelab/gridfunctionspace/gridfunctionspaceutilities.hh>
#include <dune/pdelab/gridfunctionspace/interpolate.hh>
#include <dune/pdelab/gridfunctionspace/constraints.hh>

#include "../../dune/npde/finiteelement/pk1dbasis.hh"
#include "../../dune/npde/finiteelement/pk1dconstraints.hh"

//! A function demonstrating a finite element space on a grid view
template<class GV>
void lagrange1d_basis (const GV& gv, std::size_t degree)
{
  // coordinate and result type
  typedef typename GV::Grid::ctype Coord;
  typedef double Real;

  // make the finite element map
  typedef Pk1dLocalFiniteElementMap<Coord,Real> FEM;
  FEM fem(degree);

  // make grid function space 
  typedef Dune::PDELab::NoConstraints CON;
  typedef Dune::PDELab::ISTLVectorBackend<1> VBE;
  typedef Dune::PDELab::GridFunctionSpace<GV,FEM,CON,VBE> GFS;
  GFS gfs(gv,fem);

  // make a std::vector of degree of freedom vectors
  typedef typename GFS::template VectorContainer<Real>::Type U;
  std::vector<U> u(gfs.globalSize(),U(gfs,0.0));

  // write a file for each global finite element function
  typedef Dune::PDELab::DiscreteGridFunction<GFS,U> DGF;
  std::vector<DGF *> udgf(gfs.globalSize());
  Dune::SubsamplingVTKWriter<GV> vtkwriter(gv,6);
  std::cout << "writing " << gfs.globalSize() << " basis functions" << std::endl;
  for (std::size_t i=0; i<u.size(); i++)
    {
      u[i][i] = 1.0;
      std::stringstream componentName;
      componentName << "comp_" << i;
      udgf[i] = new DGF(gfs,u[i]);
      vtkwriter.addVertexData(new Dune::PDELab::VTKGridFunctionAdapter<DGF>(*(udgf[i]),componentName.str()));
    }
  vtkwriter.write("u",Dune::VTKOptions::binaryappended);
  for (std::size_t i=0; i<udgf.size(); i++) delete udgf[i];
}

const double pi = 3.141592653589793238462643;

// a function to initialize from
template<typename GV, typename RF>
class F : public Dune::PDELab::AnalyticGridFunctionBase<
  Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,1>,
  F<GV,RF> > 
{
public:
  typedef Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,1> Traits;
  typedef Dune::PDELab::AnalyticGridFunctionBase<Traits,F<GV,RF> > B;

  F (const GV& gv) : B(gv) {}

  inline void evaluateGlobal (const typename Traits::DomainType& x, 
                              typename Traits::RangeType& y) const
  {
    y[0] = sin(3.0*pi*x[0])*sin(3.0*pi*x[0]);
    return;
  }
};

//! A function demonstrating a finite element space on a grid view
template<class GV>
void lagrange_interpolation (const GV& gv, std::size_t degree)
{
  // coordinate and result type
  typedef typename GV::Grid::ctype Coord;
  typedef double Real;

  // make the finite element map
  typedef Pk1dLocalFiniteElementMap<Coord,Real> FEM;
  FEM fem(degree);

  // make grid function space 
  typedef Dune::PDELab::NoConstraints CON;
  typedef Dune::PDELab::ISTLVectorBackend<1> VBE;
  typedef Dune::PDELab::GridFunctionSpace<GV,FEM,CON,VBE> GFS;
  GFS gfs(gv,fem);

  // make a degree of freedom vector
  typedef typename GFS::template VectorContainer<Real>::Type U;
  U u(gfs,0.0);

  // interpolate from a given grid function
  F<GV,Real> f(gv); // make analytic function object
  Dune::PDELab::interpolate(f,gfs,u);

  // write vtk file
  typedef Dune::PDELab::DiscreteGridFunction<GFS,U> DGF;
  DGF udgf(gfs,u);
  Dune::SubsamplingVTKWriter<GV> vtkwriter(gv,6);
  vtkwriter.addVertexData(new Dune::PDELab::VTKGridFunctionAdapter<DGF>(udgf,"u"));
  vtkwriter.write("interpolatedf",Dune::VTKOptions::binaryappended);
}

/** \brief boundary grid function selecting boundary conditions 
 * 0 means Neumann, 1 means Dirichlet */
template<typename GV>
class BCType : public Dune::PDELab::BoundaryGridFunctionBase<
   Dune::PDELab::BoundaryGridFunctionTraits<GV,int,1,Dune::FieldVector<int,1> >,BCType<GV> >
{
  const GV& gv;
public:
  typedef Dune::PDELab::BoundaryGridFunctionTraits<GV,int,1,Dune::FieldVector<int,1> > Traits;

  //! construct from grid view
  BCType (const GV& gv_) : gv(gv_) {}

  //! return bc type at point on intersection
  template<typename I>
  inline void evaluate (I& i, const typename Traits::DomainType& xlocal,
                        typename Traits::RangeType& y) const {  
    Dune::FieldVector<typename GV::Grid::ctype,GV::dimension> 
      x = i.geometry().global(xlocal);
    y = 0; // default is Dirichlet
    if (x[0]<1e-6)
      y = 1; // Neumann
    return;
  }

  //! get a reference to the grid view
  inline const GV& getGridView () {return gv;}
};

//! A function demonstrating a constrained finite element space on a grid view
template<class GV>
void constrained_space (const GV& gv, std::size_t degree)
{
  // coordinate and result type
  typedef typename GV::Grid::ctype Coord;
  typedef double Real;

  // make the finite element map
  typedef Pk1dLocalFiniteElementMap<Coord,Real> FEM;
  FEM fem(degree);

  // make grid function space 
  typedef Pk1dConstraints CON;
  typedef Dune::PDELab::ISTLVectorBackend<1> VBE;
  typedef Dune::PDELab::GridFunctionSpace<GV,FEM,CON,VBE> GFS;
  GFS gfs(gv,fem);

  // determine constraints for this grid function space
  typedef BCType<GV> B;    // boundary condition type
  B b(gv);
  typedef typename GFS::template ConstraintsContainer<Real>::Type CC;
  CC cc;
  Dune::PDELab::constraints(b,gfs,cc);
  std::cout << "constrained dofs=" << cc.size() << " of " << gfs.globalSize() << std::endl;

  // make a degree of freedom vector
  typedef typename GFS::template VectorContainer<Real>::Type U;
  U u(gfs,1.0);
  Dune::PDELab::set_constrained_dofs(cc,0.0,u);

  // write vtk file
  typedef Dune::PDELab::DiscreteGridFunction<GFS,U> DGF;
  DGF udgf(gfs,u);
  Dune::SubsamplingVTKWriter<GV> vtkwriter(gv,4);
  vtkwriter.addVertexData(new Dune::PDELab::VTKGridFunctionAdapter<DGF>(udgf,"u"));
  vtkwriter.write("constrainedu",Dune::VTKOptions::binaryappended);
}

int main(int argc, char **argv)
{
  // initialize MPI, finalize is done automatically on exit 
  Dune::MPIHelper::instance(argc,argv);

  // start try/catch block to get error messages from dune
  try {
    using namespace Dune;

    Dune::ConfigParser configuration;
    configuration.parseFile("pk1dglobalbasis.ini");

    // The continuous floating point type of the grid.
    typedef Dune::OneDGrid::ctype ctype;

    // Set number of vertices
    const unsigned int N = configuration.get<int>("grid.cells");
    const ctype L = 1.0;

    // Define position intervals in reference domain [0,1].
    typedef std::vector<ctype> Intervals;
    Intervals intervals(N+1);
    for(unsigned int i=0; i<N+1; ++i)
      intervals[i] = ctype(i) / ctype(N);

    // Construct grid
    Dune::OneDGrid grid(intervals);

    // call generic function
    typedef Dune::OneDGrid::LeafGridView GV;
    const GV& gv=grid.leafView();
    lagrange1d_basis(gv,3);
  }
  catch (std::exception & e) {
    std::cout << "STL ERROR: " << e.what() << std::endl;
    return 1;
  }
  catch (Dune::Exception & e) {
    std::cout << "DUNE ERROR: " << e.what() << std::endl;
    return 1;
  }
  catch (...) {
    std::cout << "Unknown ERROR" << std::endl;
    return 1;
  }

  // done
  return 0;
}
