// $Id: integration.cc 270 2010-01-14 14:47:44Z dedner $

// Dune includes
#include"config.h"           // file constructed by ./configure script
#include<dune/grid/sgrid.hh> // load sgrid definition
#include <dune/common/mpihelper.hh> // include mpi helper class 
#include <dune/common/configparser.hh> 
#include <dune/common/timer.hh> 
#include<dune/grid/yaspgrid.hh>
#include <dune/grid/io/file/vtk/subsamplingvtkwriter.hh>

#if HAVE_UG 
#include<dune/grid/uggrid.hh>
#endif
#include<dune/grid/io/file/gmshreader.hh>

//! check whether vertex is inside of triangle (0,0), (1,0), (0.5,1.0)
template<class Vertex>
bool isInsideTriangle(const Vertex & v)
{
  if(v[0] <= 0.5){
    if(v[1] <= 2*v[0]+1e-8)
      return true;
    else
      return false;
  }
  else{
    if(v[1] <= 1.0 -2*(v[0]-0.5)+1e-8)
      return true;
    else
      return false;
  }
}

#include"functors.hh"
#include"integrateentity.hh"

//! Write the VTK file.
template<class GridView, class Functor>
void writeVTKFile(const GridView & gridview, const Functor & f, std::string filename)
{
  Dune::SubsamplingVTKWriter<GridView> vtkwriter(gridview,0);
  typedef FunctorVTKFunction<typename GridView::Grid,Functor> FunctorVTKFunction;
  FunctorVTKFunction *vtk_function = new FunctorVTKFunction(f);
  typedef AllInsideFunctorVTKFunction<typename GridView::Grid,Functor> AllInsideFunctorVTKFunction;
  AllInsideFunctorVTKFunction *all_vtk_function = new AllInsideFunctorVTKFunction(f);

  vtkwriter.addVertexData(vtk_function);
  vtkwriter.addVertexData(all_vtk_function);
  vtkwriter.write(filename, Dune::VTKOptions::binaryappended);
  std::cout << "Wrote VTK file" << std::endl;
}

//! uniform refinement test
template<class Grid>
void uniformintegration (Grid& grid, const int max_level, std::string filename_base)
{
  // function to integrate
  Exp<typename Grid::ctype,Grid::dimension> f;

  // get GridView on leaf grid - type
  typedef typename Grid :: LeafGridView GridView;

  // get iterator type
  typedef typename GridView :: template Codim<0> :: Iterator LeafIterator;

  // These will hold the integral values for both modes when finished
  double old_all_inside_value(0);
  double old_some_inside_value(0);

  // loop over grid sequence
  for (int k=0; k<max_level; k++)
    {
      // get GridView instance
      GridView gridView = grid.leafView();

      // Write VTK file
      std::ostringstream ss; ss << k;
      writeVTKFile(gridView,f,filename_base + ss.str());

      // compute integral with some order
      double all_inside_value = 0.0;	  
      double some_inside_value = 0.0;
      LeafIterator eendit = gridView.template end<0>();
      for (LeafIterator it = gridView.template begin<0>(); it!=eendit; ++it){
            
        const int corners = it->geometry().corners();
        bool all_inside = true;
        bool some_inside = false;
        for(int i=0; i<corners; ++i){
          const bool vertex_is_inside = isInsideTriangle(it->geometry().corner(i));
          all_inside = all_inside   && vertex_is_inside;
          some_inside = some_inside || vertex_is_inside;
        }
        const double cell_integral = integrateentity(it,f,1);
        if(all_inside)
          all_inside_value += cell_integral;
        if(some_inside)
          some_inside_value += cell_integral;
      }

      // print result and error estimate
      std::cout << "elements=" 
                << std::setw(8) << std::right
                << grid.size(0) << std::endl
                << "all inside integral=" 
                << std::scientific << std::setprecision(12) 
                << all_inside_value 
                << " error=" << std::abs(all_inside_value-old_all_inside_value) 
                << std::endl
                << "some inside integral=" 
                << std::scientific << std::setprecision(12) 
                << some_inside_value 
                << " error=" << std::abs(some_inside_value-old_some_inside_value) 

                << std::endl;

      // save value of integral
      old_all_inside_value = all_inside_value;
      old_some_inside_value = some_inside_value;

      // refine all elements
      if(k<max_level-1)
        grid.globalRefine(1);
    }
}

int main(int argc, char **argv)
{
  // initialize MPI, finalize is done automatically on exit 
  Dune::MPIHelper::instance(argc,argv);

  // start try/catch block to get error messages from dune
  try {
    using namespace Dune;

    Dune::ConfigParser configuration;
    configuration.parseFile("loesung01.ini");
    
#if HAVE_UG
    std::cout << "\nUnstructured grid integration:" << std::endl;
    { // UG grid example
      Dune::Timer timer;
      typedef Dune::UGGrid<2> GridType;
      GridType grid;

      typedef std::vector<int> GmshIndexMap;
      GmshIndexMap boundary_index_map;
      GmshIndexMap element_index_map;

      std::string grid_file = configuration.get<std::string>("grid.filename");
      Dune::GmshReader<GridType> gmsh_reader;
      gmsh_reader.read(grid,grid_file,boundary_index_map,element_index_map,true,false);

      // The prefix of all VTK files
      const std::string filename_base = "ug_functor_";

      // integrate and compute error with extrapolation
      uniformintegration( grid, configuration.get<int>("grid.max_level"), filename_base );
      std::cout << "Time for unstructured integration: " << timer.elapsed() << std::endl;
    }
#endif

    std::cout << "\nStructured grid integration:" << std::endl;
    { // YASP grid example
      Dune::Timer timer;
      Dune::FieldVector<double,2> L(1.0);
      Dune::FieldVector<int,2> N(1);
      Dune::FieldVector<bool,2> B(false);
      Dune::YaspGrid<2> grid(L,N,B,0);

      // The prefix of all VTK files
      const std::string filename_base = "yasp_functor_";

      // integrate and compute error with extrapolation
      uniformintegration( grid, configuration.get<int>("grid.max_level") , filename_base);

      std::cout << "Time for structured integration: " << timer.elapsed() << std::endl;
    }


  }
  catch (std::exception & e) {
    std::cout << "STL ERROR: " << e.what() << std::endl;
    return 1;
  }
  catch (Dune::Exception & e) {
    std::cout << "DUNE ERROR: " << e.what() << std::endl;
    return 1;
  }
  catch (...) {
    std::cout << "Unknown ERROR" << std::endl;
    return 1;
  }

  // done
  return 0;
}
