#ifndef STRING_J_ASSEMBLER_HH
#define STRING_J_ASSEMBLER_HH

template <typename Rhs, typename ForceFunctor>
void assembleRhs(Rhs & rhs, const Rhs & vertices, const ForceFunctor & f, const int dim)
{
  typedef typename ForceFunctor::Coord Coordinate;
  const unsigned int N = vertices.size() / dim - 1;
  // Reset right hand side
  rhs = 0;
  for(int i=0; i<dim; ++i){
    rhs[i] = vertices[i];
    rhs[N*dim + i] = vertices[N*dim + i];
  }
  for(int i=1; i<N; ++i){
    Coordinate c(0);
    for(int j=0; j<dim; ++j)
      c[j]=vertices[dim*i+j];
    const Coordinate force = f(c);
    for(int j=0; j<dim; ++j)
      rhs[dim*i + j] = force[j];
  }


}

template<typename Matrix, typename ScalarData, int dim>
void assembleMatrix(Matrix & A, const ScalarData & u, 
                    const ScalarData & kappa, const ScalarData & lengths)
{

  typedef typename ScalarData::value_type ctype;

  const unsigned int N = u.size() / dim - 1;

  // Compute absolute differences
  ScalarData abs_diffs(N);
  abs_diffs=0;
  for(unsigned int i=0; i<N; ++i){
    for(int j=0; j<dim; ++j){
      const ctype diff = u[(i+1)*dim + j] - u[i*dim + j];
      abs_diffs[i] += diff * diff;
    }
    abs_diffs[i] = std::sqrt(abs_diffs[i]);
  }

  // Get row iterators
  typename Matrix::RowIterator rit = A.begin();

  // Fill the first dim rows with trivial problems
  for(int i=0; i<dim; ++i){
    A[i][i] = 1.0;
    ++rit;
  }

  // Iterate rows
  while(true){
    const int ri = rit.index();
    const int coord_index = ri / dim;

    // Do not process the last dim lines
    if(coord_index == N)
      break;

    const ctype nl_factor_km1 
      = kappa[coord_index-1]*(abs_diffs[coord_index-1]-lengths[coord_index-1]) / abs_diffs[coord_index-1];
    const ctype nl_factor_kp1 
      = kappa[coord_index]*(abs_diffs[coord_index]-lengths[coord_index]) / abs_diffs[coord_index];
    ctype factors[3];
    factors[0]=-nl_factor_km1;
    factors[1]=nl_factor_km1 + nl_factor_kp1;
    factors[2]=-nl_factor_kp1;
    
  // Get column iterators
    typename Matrix::ColIterator cit = rit->begin();
    typename Matrix::ColIterator ecit = rit->end();

    // Iterate columns
    for(; cit != ecit; ++cit){
      const int ci = cit.index();
      const int factor_index = (ci - ri) / dim + 1 ;
      assert(factor_index >= 0);
      assert(factor_index <= 2);
      *cit = factors[factor_index];
    }

    ++rit;
  }

  // Fill the last dim rows with trivial problems
  for(int i=0; i<dim; ++i){
    A[N*dim+i][N*dim+i] = 1.0;
  }

}

#endif
