// $Id: integration.cc 270 2010-01-14 14:47:44Z dedner $

// Dune includes
#include"config.h"           // file constructed by ./configure script
#include<dune/grid/sgrid.hh> // load sgrid definition
#include <dune/common/mpihelper.hh> // include mpi helper class 
#include <dune/common/configparser.hh> // include mpi helper class 
#include<dune/grid/onedgrid.hh>
#include <dune/grid/io/file/vtk/subsamplingvtkwriter.hh>
#if HAVE_UG 
#include<dune/grid/uggrid.hh>
#endif
#include<dune/grid/io/file/gmshreader.hh>
#include<sstream>

#include"functors.hh"
#include"integrateentity.hh"

//! Write the VTK file.
template<class GridView, class ReferenceFunctor, class ApproxFunctor>
void writeVTKFile(const GridView & gridview, const ReferenceFunctor & rf, 
                  const ApproxFunctor & af, std::string filename)
{
  Dune::SubsamplingVTKWriter<GridView> vtkwriter(gridview,0);    
  typedef FunctorVTKFunction<typename GridView::Grid,ApproxFunctor> VTKApproxFunctor;
  VTKApproxFunctor *vtk_af = new VTKApproxFunctor(af,"Approximation");
  typedef FunctorVTKFunction<typename GridView::Grid,ReferenceFunctor> VTKF;
  VTKF *vtk_rf = new VTKF(rf,"Reference");
  vtkwriter.addVertexData(vtk_af);
  vtkwriter.addVertexData(vtk_rf);
  vtkwriter.write(filename, Dune::VTKOptions::binaryappended);
  std::cout << "Wrote VTK file" << std::endl;
}

//! uniform refinement test
template<class Grid, class Functor>
double uniformintegration (Grid& grid, const Functor & f, const int qorder, 
                           const int max_level, const double accuracy)
{
  // get GridView on leaf grid - type
  typedef typename Grid :: LevelGridView GridView;

  // loop over grid sequence
  double oldvalue=0;
  const std::string filename_base = "functor_";
  for (int k=0; k<max_level; k++)
    {
      // get GridView instance
      GridView gridView = grid.levelView(k);

      // Write VTK file
      std::ostringstream ss; ss << k;

      // compute integral with some order
      const double value = integrategrid(gridView,f,qorder);

      const double diff = std::abs(value-oldvalue);
      // print result and error estimate
      std::cout << "elements=" 
                << std::setw(8) << std::right
                << gridView.size(0)
                << " integral=" 
                << std::scientific << std::setprecision(12) 
                << value 
                << " difference=" << diff
                << std::endl;

      if(diff < accuracy)
        return value;

      // save value of integral
      oldvalue=value;

    }

  std::cerr << "Could not achieve sufficient quadrature accuracy." << std::endl;
  exit(1);
  return 0;
}

int main(int argc, char **argv)
{
  // initialize MPI, finalize is done automatically on exit 
  Dune::MPIHelper::instance(argc,argv);

  Dune::ConfigParser configuration;
  configuration.parseFile("uebung03.ini");

  // start try/catch block to get error messages from dune
  try {
    using namespace Dune;

    const int N = configuration.get<int>("grid.elements");
    typedef Dune::OneDGrid Grid;
    Grid grid(N,0,1);


    // function to integrate
    typedef Phi<Grid::ctype, Grid::dimension> F;
    F f;
    const double t0 = configuration.get<double>("setup.t0");
    f.setTime(t0);

    // integrate and compute error with extrapolation
    const int quad_order = configuration.get<int>("integration.order");
    const int max_level = configuration.get<int>("integration.max_level");
    const double accuracy = configuration.get<double>("integration.accuracy");

    // Create grid levels
    for(int l=0; l<max_level; ++l)
      grid.globalRefine(1);

    typedef SinNPiX<Grid::ctype> SinN;
    typedef CosNPiX<Grid::ctype> CosN;
    const int modes = configuration.get<int>("fourier.modes");
    std::vector<double> a_n(modes);
    std::vector<double> b_n(modes);
    for(int n=0; n<modes; ++n){
      SinN sin_n(n);
      ProductFunctor<SinN,F> sin_n_f(sin_n,f);
      a_n[n] =  2 * uniformintegration( grid, sin_n_f, quad_order, max_level, accuracy);
      std::cout << "Coefficient a_" << n << " : " << a_n[n] << std::endl;

      CosN cos_n(n);
      ProductFunctor<CosN,F> cos_n_f(cos_n,f);
      b_n[n] = 2 * uniformintegration( grid, cos_n_f, quad_order, max_level, accuracy );
      std::cout << "Coefficient b_" << n << " : " << b_n[n] << std::endl;
    }

    typedef FourierSeries<Grid::ctype> FourierSeriesFunctor;
    FourierSeriesFunctor fsf(a_n,b_n);

    typedef FourierSeriesEvolution<Grid::ctype> FourierSeriesEvolutionFunctor;
    FourierSeriesEvolutionFunctor fsef(fsf);

    const double tmax = configuration.get<double>("setup.tmax");
    const double dt = configuration.get<double>("setup.dt");
    int step=0;
    for(double t=t0; t<tmax; t+=dt,step++){
      fsef.setTime(t-t0);
      f.setTime(t);

      SquareDiffFunctor<FourierSeriesEvolutionFunctor,F> l2_functor(fsef,f);
      const double l2_error = 
        std::sqrt(uniformintegration( grid, l2_functor, quad_order, max_level, accuracy));
      std::cout << "Time: " << t << "  L2-Error: " << l2_error << std::endl;

      std::ostringstream postfix;postfix << step;
      writeVTKFile(grid.leafView(),f,fsef,std::string("evolution_") + postfix.str());
    }
    
  }
  catch (std::exception & e) {
    std::cout << "STL ERROR: " << e.what() << std::endl;
    return 1;
  }
  catch (Dune::Exception & e) {
    std::cout << "DUNE ERROR: " << e.what() << std::endl;
    return 1;
  }
  catch (...) {
    std::cout << "Unknown ERROR" << std::endl;
    return 1;
  }

  // done
  return 0;
}
