#ifndef __DUNE_GRID_HOWTO_FUNCTORS_HH__
#define __DUNE_GRID_HOWTO_FUNCTORS_HH__

#include <dune/common/fvector.hh>

// A functor representing the fundamental solution of the heat
// transport equation.
template<typename ct, int dim>
class Phi {
public:
  Phi () {midpoint = 0.5; t = 0;}
  double operator() (const Dune::FieldVector<ct,dim>& x) const
  {
	Dune::FieldVector<ct,dim> y(x);
	y -= midpoint;
        const double ft = std::sqrt(4*t);
        double prefactor = 1.0;
        for(int d=0; d<dim; ++d) prefactor *= ft;
        return 1/prefactor*std::exp(-(y*y)/(4*t));
  }
  void setTime(const double t_){t=t_;}
private:
  Dune::FieldVector<ct,dim> midpoint;
  double t;
};

// Functor for evaluation of PK simplex function
template<typename LocalFunction>
class LocalFunctor{
private:
  typedef typename LocalFunction::Traits::LocalBasisType::Traits::RangeType RT;
  typedef typename LocalFunction::Traits::LocalBasisType::Traits::RangeFieldType RFT;
  typedef typename LocalFunction::Traits::LocalBasisType::Traits::DomainType DT;
  typedef std::vector<RFT> Coefficients;

  const LocalFunction local_function;
  const Coefficients coeffs;

public:
  LocalFunctor(const LocalFunction & l_, const Coefficients & coeffs_) 
    : local_function(l_), coeffs(coeffs_)
  {}

  RT operator() (const DT & x) const
  {
    RT value(0);
    std::vector<RT> values;
    local_function.localBasis().evaluateFunction(x,values);
    for(unsigned int i=0; i<values.size(); ++i)
      value += coeffs[i] * values[i];
    return value;
  }
};

template <int k, class RangeType>
class MonomFunctor{
private:
  static const int dim = 2;
  typedef Dune::FieldVector<RangeType,dim> Exponent;

  std::vector< Exponent > exponents;
  int basis_index;

  RangeType pow(const RangeType & x, int e) const
  {
    RangeType v(1.0);
    for(int i=0; i<e; ++i)
      v *= x;
    return v;
  }

public:
  MonomFunctor()
  {
    for(int p=0; p<=k; ++p){
      for(int yp=0; yp<=p; ++yp){
        const int xp = p - yp;
        Exponent ex; 
        ex[0] = xp;
        ex[1] = yp;
        std::cout << "EXP: " << ex[0] << " " << ex[1] << std::endl;
        exponents.push_back(ex);
      }
    }

    basis_index = 0;
  }

  int size() {return exponents.size(); }

  void setIndex(const int i)
  {
    basis_index = i;
  }

  template<class Coordinate>
  RangeType operator()(const Coordinate & x) const
  {
    RangeType v(0);
    const Exponent & exponent = exponents[basis_index];
    v = pow(x[0],exponent[0]) * pow(x[1],exponent[1]);
    return v;
  }

};

// Functor for evaluation of PK simplex function
template<typename LocalFunction>
class LocalJacobianSquareFunctor{
private:
  typedef typename LocalFunction::Traits::LocalBasisType::Traits::RangeType RT;
  typedef typename LocalFunction::Traits::LocalBasisType::Traits::JacobianType JT;
  typedef typename LocalFunction::Traits::LocalBasisType::Traits::RangeFieldType RFT;
  typedef typename LocalFunction::Traits::LocalBasisType::Traits::DomainType DT;
  typedef std::vector<RFT> Coefficients;

  const LocalFunction local_function;
  const Coefficients coeffs;

public:
  LocalJacobianSquareFunctor(const LocalFunction & l_, const Coefficients & coeffs_) 
    : local_function(l_), coeffs(coeffs_)
  {}

  RT operator() (const DT & x) const
  {
    JT value(0);
    std::vector<JT> values;
    local_function.localBasis().evaluateJacobian(x,values);
    for(unsigned int i=0; i<values.size(); ++i){
      value.axpy(coeffs[i],values[i]);
    }
    RT square(0);
    for(unsigned int i=0; i<value.M(); ++i){
      assert(value.M() == 2);
      assert(value.exists(0,i));
      square += value[0][i] * value[0][i];
    }
    return square;
  }
};

template< typename F1, typename F2>
class ProductFunctor
{
  const F1 & f1;
  const F2 & f2;
public:
  ProductFunctor(const F1 & f1_, const F2 & f2_)
    : f1(f1_), f2(f2_)
  {}

  template<typename Coordinate>
  double operator() (const Coordinate& x) const
  {
    return f1(x)*f2(x);
  }
};


/**
   This function may be used by the VTKWriter object to evaluate the
   functor object at a given local coordinate.
 */
template<class Grid, class Functor>
class FunctorVTKFunction : public Dune::VTKFunction<Grid>
{
private:

  // Some types which are exported by the base class
  typedef Dune::VTKFunction<Grid> Base;
  typedef typename Base::Entity Entity;
  typedef typename Base::ctype ctype;
  enum { dim =  Base::dim };

  // Constant reference to the functor object.
  const Functor & f;
  const std::string data_name;
public:

  FunctorVTKFunction(const Functor & f_, std::string name_) : f(f_), data_name(name_)
  {}

  //! The number of components of this function. This is necessary as
  //! the VTKWriter would also allow the visualization of vector
  //! valued functions.
  int ncomps() const {return 1;}

  //! Evaluate the functor at a given local coordinate
  double evaluate(int comp, const Entity & e, const Dune::FieldVector<ctype,dim> & xi) const
  {
    typedef typename Entity::Geometry::GlobalCoordinate GlobalCoordinate;
    const GlobalCoordinate c = e.geometry().global(xi);
    const double f_value = f(c);
    return f_value;
  }

  //! This name will later be displayed as a label for the visualized
  //! data in paraview.
  std::string name() const
  {
    return data_name;
  }
};


#endif // __DUNE_GRID_HOWTO_FUNCTORS_HH__
