// $Id: integration.cc 270 2010-01-14 14:47:44Z dedner $

// Dune includes
#include"config.h"           // file constructed by ./configure script
#include<dune/grid/sgrid.hh> // load sgrid definition
#include <dune/common/mpihelper.hh> // include mpi helper class 
#include <dune/common/configparser.hh> // include mpi helper class 
#include<dune/grid/onedgrid.hh>
#include <dune/grid/io/file/vtk/subsamplingvtkwriter.hh>
#if HAVE_UG 
#include<dune/grid/uggrid.hh>
#endif
#include<dune/grid/io/file/gmshreader.hh>
#include<sstream>

#include <dune/localfunctions/monom.hh>
#include <dune/localfunctions/lagrange/pk.hh>
#include"functors.hh"
#include"integrateentity.hh"

//! Write the VTK file.
template<class GridView, class Functor>
void writeVTKFile(const GridView & gridview, const Functor & f, std::string filename)
{
  Dune::SubsamplingVTKWriter<GridView> vtkwriter(gridview,0);    
  typedef FunctorVTKFunction<typename GridView::Grid,Functor> VTKFunctor;
  VTKFunctor *vtk_f = new VTKFunctor(f,"Functor");
  vtkwriter.addVertexData(vtk_f);
  vtkwriter.write(filename, Dune::VTKOptions::binaryappended);
  std::cout << "Wrote VTK file" << std::endl;
}

//! uniform refinement test
template<class Grid, class Functor>
double uniformintegration (Grid& grid, const Functor & f, const int qorder, 
                           const int max_level, const double accuracy)
{
  // get GridView on leaf grid - type
  typedef typename Grid :: LevelGridView GridView;

  // loop over grid sequence
  double oldvalue=0;
  const std::string filename_base = "functor_";
  for (int k=0; k<max_level; k++)
    {
      // get GridView instance
      GridView gridView = grid.levelView(k);

      // Write VTK file
      std::ostringstream ss; ss << k;

      // compute integral with some order
      const double value = integrategrid(gridView,f,qorder);

      const double diff = std::abs(value-oldvalue);
      // print result and error estimate
      // std::cout << "elements=" 
      //           << std::setw(8) << std::right
      //           << gridView.size(0)
      //           << " integral=" 
      //           << std::scientific << std::setprecision(12) 
      //           << value 
      //           << " difference=" << diff
      //           << std::endl;

      if(diff < accuracy)
        return value;

      // save value of integral
      oldvalue=value;

    }

  std::cerr << "Could not achieve sufficient quadrature accuracy." << std::endl;
  exit(1);
  return 0;
}

template<typename Grid, int dim>
Grid* createSingleSimplexGrid()
{
  Dune::GridFactory<Grid> grid_factory;
  typedef typename Grid::template Codim<0>::Geometry::LocalCoordinate Coordinate;
  Coordinate vertex;

  vertex = 0;
  grid_factory.insertVertex(vertex);

  for(int i=0; i<dim; ++i){
    vertex = 0;
    vertex[i] = 1.0;
    grid_factory.insertVertex(vertex);
  }
  
  const int n_vertices = dim + 1;
  std::vector<unsigned int> indices(n_vertices);
  for(int i=0; i<n_vertices; ++i) 
    indices[i]=i;

  Dune::GeometryType simplex_type(Dune::GeometryType::simplex,dim);
  grid_factory.insertElement(simplex_type,indices);

  return grid_factory.createGrid();
}

void output_result
(const std::vector<double> & l2_norms2,
 const std::vector<double> & h1_snorms2,
 const std::vector<double> & square_integrals)
{
  double max = std::numeric_limits<double>::min();
  double min = std::numeric_limits<double>::max();
  double avg = 0;
  for(unsigned int i=0; i<l2_norms2.size(); ++i){
    const double ratio = (square_integrals[i] + h1_snorms2[i]) / l2_norms2[i];
    max = std::max(ratio,max);
    min = std::min(ratio,min);
    avg += ratio;
  }
  avg /= double(l2_norms2.size());
  std::cout << "Max: " << max << " Min: " << min << " Avg: " << avg << std::endl;
}


int main(int argc, char **argv)
{
  // initialize MPI, finalize is done automatically on exit 
  Dune::MPIHelper::instance(argc,argv);

  Dune::ConfigParser configuration;
  configuration.parseFile("uebung05.ini");

  // start try/catch block to get error messages from dune
  try {
    using namespace Dune;

    const int dim = 2;
    const int k = 3;
    typedef Dune::UGGrid<dim> Grid;
 
    Grid * pgrid = createSingleSimplexGrid<Grid,dim>();
    Grid & grid = *pgrid;
    
    // integrate and compute error with extrapolation
    const int quad_order = configuration.get<int>("integration.order");
    const double max_level = configuration.get<double>("integration.max_level");
    const double accuracy = configuration.get<double>("integration.accuracy");

    // Create grid levels
    for(int l=0; l<max_level; ++l)
      grid.globalRefine(1);

    typedef Grid::ctype DF;
    typedef double RF;
    typedef Dune::PkLocalFiniteElement<DF,RF,dim,k> PKFunction;
    typedef LocalFunctor<PKFunction> PKFunctor;
    typedef LocalJacobianSquareFunctor<PKFunction> PKJacobianSquareFunctor;
    typedef PKFunction::Traits::LocalBasisType::Traits::RangeFieldType LocalRangeField;

    PKFunction local_function;
    std::vector<LocalRangeField> coefficients(local_function.localBasis().size());

    typedef MonomFunctor<k,double> MonomFunctor;
    MonomFunctor pk_monoms;

    const unsigned int coeff_size = Dune::MonomImp::Size<dim,k>::val;
    Dune::FieldMatrix<double,coeff_size,coeff_size> A(0);

    assert(coefficients.size() == coeff_size);

    for(unsigned int c=0; c<coefficients.size(); ++c){
      if(c>0) coefficients[c-1] = 0;
      coefficients[c] = 1;
      PKFunctor local_functor(local_function, coefficients);

      for(unsigned int d=0; d<coeff_size; ++d){

        pk_monoms.setIndex(d);
        ProductFunctor<MonomFunctor,PKFunctor> product(pk_monoms,local_functor);

        if(c==0){
          std::ostringstream ss; ss << d; 
          std::string filename = std::string("monom_functor_") + ss.str();
          writeVTKFile(grid.leafView(),pk_monoms,filename);
        }

        A[d][c] = uniformintegration (grid, product, quad_order, max_level, accuracy);
      }

      std::ostringstream ss; ss << c; 
      std::string filename = std::string("pk_functor_") + ss.str();
      writeVTKFile(grid.leafView(),local_functor,filename);
    }

    std::cout << "Matrix: " << std::endl << A << std::endl;
    Dune::FieldMatrix<double,coeff_size,coeff_size> unit(A);
    unit.invert(); 
    std::cout << "Inverse: " << std::endl << unit << std::endl;
    unit.rightmultiply(A);
    std::cout << "Unit: " << std::endl << std::fixed << unit << std::endl;

    delete pgrid;
  }
  catch (std::exception & e) {
    std::cout << "STL ERROR: " << e.what() << std::endl;
    return 1;
  }
  catch (Dune::Exception & e) {
    std::cout << "DUNE ERROR: " << e.what() << std::endl;
    return 1;
  }
  catch (...) {
    std::cout << "Unknown ERROR" << std::endl;
    return 1;
  }

  // done
  return 0;
}
