// $Id: integration.cc 270 2010-01-14 14:47:44Z dedner $

// Dune includes
#include"config.h"           // file constructed by ./configure script
#include<dune/grid/sgrid.hh> // load sgrid definition
#include <dune/common/mpihelper.hh> // include mpi helper class 
#include<dune/grid/yaspgrid.hh>
#include <dune/grid/io/file/vtk/subsamplingvtkwriter.hh>
#if HAVE_UG 
#include<dune/grid/uggrid.hh>
#endif
#include<dune/grid/io/file/gmshreader.hh>

#include"functors.hh"
#include"integrateentity.hh"

//! Write the VTK file.
template<class GridView, class Functor>
void writeVTKFile(const GridView & gridview, const Functor & f, std::string filename)
{
  Dune::SubsamplingVTKWriter<GridView> vtkwriter(gridview,0);
  typedef FunctorVTKFunction<typename GridView::Grid,Functor> FunctorVTKFunction;
  FunctorVTKFunction *vtk_function = new FunctorVTKFunction(f);
  vtkwriter.addVertexData(vtk_function);
  vtkwriter.write(filename, Dune::VTKOptions::binaryappended);
  std::cout << "Wrote VTK file" << std::endl;
}

//! uniform refinement test
template<class Grid>
void uniformintegration (Grid& grid)
{
  // function to integrate
  Exp<typename Grid::ctype,Grid::dimension> f;

  // get GridView on leaf grid - type
  typedef typename Grid :: LeafGridView GridView;

  // get iterator type
  typedef typename GridView :: template Codim<0> :: Iterator LeafIterator;

  // loop over grid sequence
  double oldvalue=0;
  const int max_level = 9;
  const std::string filename_base = "functor_";
  for (int k=0; k<max_level; k++)
    {
      // get GridView instance
      GridView gridView = grid.leafView();

      // Write VTK file
      std::ostringstream ss; ss << k;
      writeVTKFile(gridView,f,filename_base + ss.str());

      // compute integral with some order
      double value = 0.0;
      LeafIterator eendit = gridView.template end<0>();
      for (LeafIterator it = gridView.template begin<0>(); it!=eendit; ++it)
        value += integrateentity(it,f,1);        /*@\label{ic:call}@*/

      // print result and error estimate
      std::cout << "elements=" 
                << std::setw(8) << std::right
                << grid.size(0)
                << " integral=" 
                << std::scientific << std::setprecision(12) 
                << value 
                << " difference=" << std::abs(value-oldvalue) 
                << std::endl;

      // save value of integral
      oldvalue=value;

      // refine all elements
      if(k < max_level - 1)
        grid.globalRefine(1);
    }
}

int main(int argc, char **argv)
{
  // initialize MPI, finalize is done automatically on exit 
  Dune::MPIHelper::instance(argc,argv);

  // start try/catch block to get error messages from dune
  try {
    using namespace Dune;

    Dune::FieldVector<double,2> L(1.0);
    Dune::FieldVector<int,2> N(1);
    Dune::FieldVector<bool,2> B(false);
    Dune::YaspGrid<2> grid(L,N,B,0);

    // integrate and compute error with extrapolation
    uniformintegration( grid );
  }
  catch (std::exception & e) {
    std::cout << "STL ERROR: " << e.what() << std::endl;
    return 1;
  }
  catch (Dune::Exception & e) {
    std::cout << "DUNE ERROR: " << e.what() << std::endl;
    return 1;
  }
  catch (...) {
    std::cout << "Unknown ERROR" << std::endl;
    return 1;
  }

  // done
  return 0;
}
