// $Id: integration.cc 270 2010-01-14 14:47:44Z dedner $

// Dune includes
#include"config.h"           // file constructed by ./configure script
#include<dune/grid/sgrid.hh> // load sgrid definition
#include <dune/common/mpihelper.hh> // include mpi helper class 
#include <dune/common/configparser.hh> // include mpi helper class 
#include<dune/grid/onedgrid.hh>
#include <dune/grid/io/file/vtk/subsamplingvtkwriter.hh>
#if HAVE_UG 
#include<dune/grid/uggrid.hh>
#endif
#include<dune/grid/io/file/gmshreader.hh>
#include<sstream>

#include <dune/localfunctions/monom.hh>
#include <dune/localfunctions/lagrange/pk.hh>
#include"functors.hh"
#include"integrateentity.hh"

//! Write the VTK file.
template<class GridView, class Functor>
void writeVTKFile(const GridView & gridview, const Functor & f, std::string filename)
{
  Dune::SubsamplingVTKWriter<GridView> vtkwriter(gridview,0);    
  typedef FunctorVTKFunction<typename GridView::Grid,Functor> VTKFunctor;
  VTKFunctor *vtk_f = new VTKFunctor(f,"Functor");
  vtkwriter.addVertexData(vtk_f);
  vtkwriter.write(filename, Dune::VTKOptions::binaryappended);
  std::cout << "Wrote VTK file" << std::endl;
}

//! uniform refinement test
template<class Grid, class Functor>
double uniformintegration (Grid& grid, const Functor & f, const int qorder, 
                           const int max_level, const double accuracy)
{
  // get GridView on leaf grid - type
  typedef typename Grid :: LevelGridView GridView;

  // loop over grid sequence
  double oldvalue=0;
  const std::string filename_base = "functor_";
  for (int k=0; k<max_level; k++)
    {
      // get GridView instance
      GridView gridView = grid.levelView(k);

      // Write VTK file
      std::ostringstream ss; ss << k;

      // compute integral with some order
      const double value = integrategrid(gridView,f,qorder);

      const double diff = std::abs(value-oldvalue);
      // print result and error estimate
      std::cout << "elements=" 
                << std::setw(8) << std::right
                << gridView.size(0)
                << " integral=" 
                << std::scientific << std::setprecision(12) 
                << value 
                << " difference=" << diff
                << std::endl;

      if(diff < accuracy)
        return value;

      // save value of integral
      oldvalue=value;

    }

  std::cerr << "Could not achieve sufficient quadrature accuracy." << std::endl;
  exit(1);
  return 0;
}

template<typename Grid, int dim>
Grid* createSingleSimplexGrid()
{
  Dune::GridFactory<Grid> grid_factory;
  typedef typename Grid::template Codim<0>::Geometry::LocalCoordinate Coordinate;
  Coordinate vertex;

  vertex = 0;
  grid_factory.insertVertex(vertex);

  for(int i=0; i<dim; ++i){
    vertex = 0;
    vertex[i] = 1.0;
    grid_factory.insertVertex(vertex);
  }
  
  const int n_vertices = dim + 1;
  std::vector<unsigned int> indices(n_vertices);
  for(int i=0; i<n_vertices; ++i) 
    indices[i]=i;

  Dune::GeometryType simplex_type(Dune::GeometryType::simplex,dim);
  grid_factory.insertElement(simplex_type,indices);

  return grid_factory.createGrid();
}

int main(int argc, char **argv)
{
  // initialize MPI, finalize is done automatically on exit 
  Dune::MPIHelper::instance(argc,argv);

  Dune::ConfigParser configuration;
  configuration.parseFile("uebung05.ini");

  // start try/catch block to get error messages from dune
  try {
    using namespace Dune;

    // The dimension of the domain
    const int dim = 2;

    // The polynomial order of the Pk basis
    const int k = 3;

    // Type  of the simplex grid
    typedef Dune::UGGrid<dim> Grid;

    // Create a grid consisting of a single simplex embedded in the
    // unit cube.
    Grid * pgrid = createSingleSimplexGrid<Grid,dim>();
    Grid & grid = *pgrid;
    
    // integrate and compute error with extrapolation
    const int quad_order = configuration.get<int>("integration.order");
    const double max_level = configuration.get<double>("integration.max_level");
    const double accuracy = configuration.get<double>("integration.accuracy");

    // Create grid levels.
    for(int l=0; l<max_level; ++l)
      grid.globalRefine(1);

    typedef Grid::ctype DF;
    typedef double RF;
    typedef Dune::PkLocalFiniteElement<DF,RF,dim,k> PKBasisFunctions;
    typedef LocalFunctor<PKBasisFunctions> PKFunctor;
    typedef PKBasisFunctions::Traits::LocalBasisType::Traits::RangeFieldType LocalRangeField;

    // The PK local basis functions and their coefficients
    PKBasisFunctions local_functions;

    // A matrix with dimension equal to the number of basis functions (coefficients)
    const unsigned int coeff_size = Dune::MonomImp::Size<dim,k>::val;
    Dune::FieldMatrix<double,coeff_size,coeff_size> A(0);
    assert(local_functions.localBasis().size() == coeff_size);

    // Example evaluation of Pk local basis functions at coordinates (0.5,0.5)
    typedef PKBasisFunctions::Traits::LocalBasisType::Traits::RangeType RT;
    typedef PKBasisFunctions::Traits::LocalBasisType::Traits::RangeFieldType RFT;
    typedef PKBasisFunctions::Traits::LocalBasisType::Traits::DomainType DT;
    typedef PKBasisFunctions::Traits::LocalBasisType::Traits::JacobianType JT;
    typedef std::vector<RFT> Coefficients;
    DT domain_coordinates(0.25);
    std::vector<RT> values;
    local_functions.localBasis().evaluateFunction(domain_coordinates,values);
    for(unsigned int i=0;i<values.size();++i)
      std::cout << "Pk basis function " << i << " has value " << values[i] 
                << " at point " << domain_coordinates << std::endl;

    std::vector<JT> jacobian_values;
    local_functions.localBasis().evaluateJacobian(domain_coordinates,jacobian_values);
    for(unsigned int i=0;i<jacobian_values.size();++i)
      std::cout << "Pk basis function " << i << " has derivatives " << std::endl 
                << jacobian_values[i] << " at point " << domain_coordinates << std::endl;


    // After you have implemented the PKFunctor correctly, this
    // snippet of code should write VTK files for each basis function
    std::vector<LocalRangeField> coefficients(coeff_size,0);
    for(unsigned int c=0; c<coefficients.size(); ++c){
      if(c>0) coefficients[c-1] = 0;
      coefficients[c] = 1;
      PKFunctor local_functor(local_functions, coefficients);

      std::ostringstream ss; ss << c; 
      std::string filename = std::string("pk_functor_") + ss.str();
      writeVTKFile(grid.leafView(),local_functor,filename);
    }

    // In the code below, we check, whether the matrix is singular
    
    // std::cout << "Matrix: " << std::endl << A << std::endl;
    // Dune::FieldMatrix<double,coeff_size,coeff_size> unit(A);
    // unit.invert(); 
    // std::cout << "Inverse: " << std::endl << unit << std::endl;
    // unit.rightmultiply(A);
    // std::cout << "Unit: " << std::endl << std::fixed << unit << std::endl;

    delete pgrid;
  }
  catch (std::exception & e) {
    std::cout << "STL ERROR: " << e.what() << std::endl;
    return 1;
  }
  catch (Dune::Exception & e) {
    std::cout << "DUNE ERROR: " << e.what() << std::endl;
    return 1;
  }
  catch (...) {
    std::cout << "Unknown ERROR" << std::endl;
    return 1;
  }

  // done
  return 0;
}
