#ifndef __DUNE_GRID_HOWTO_FUNCTORS_HH__
#define __DUNE_GRID_HOWTO_FUNCTORS_HH__

#include <dune/common/fvector.hh>

// a gaussian function
template<typename GV, typename RF>
class Harmonic : 
  public Dune::PDELab::AnalyticGridFunctionBase<
  Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,1>,
  Harmonic<GV,RF> > 
{
public:
  typedef Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,1> Traits;
  typedef Dune::PDELab::AnalyticGridFunctionBase<Traits,Harmonic<GV,RF> > Base;
  typedef typename Traits::DomainType DT;
  typedef typename Traits::RangeType RT;
  

  Harmonic (const GV& gv) 
    : Base(gv)  {}

  inline void evaluateGlobal (const DT & x, RT & y) const
  {
    DT c(x);
    y[0] = 1/(x[0]+0.5) + 1/(x[1]+0.5);
  }
};


template<typename Grid, typename DGF>
class GridLevelFunction
{
private:
  const DGF & dgf;
  int level;
  typedef typename DGF::Traits::DomainType DT;
  typedef typename DGF::Traits::RangeType RT;
  typedef typename Grid::template Codim<0>::Entity Entity;
  typedef typename Grid::template Codim<0>::EntityPointer EntityPointer;

public:
  
  typedef typename DGF::Traits Traits;
  
  GridLevelFunction(const DGF & dgf_, const int level_)
    : dgf(dgf_), level(level_)
  {}

  void evaluate(const Entity & e, const DT & lx, RT & y) const
  {
    EntityPointer ep(e);
    while(ep->level() != level)
      ep = ep->father();
    const DT gx(e.geometry().global(lx));
    const DT cx(ep->geometry().local(gx));
    dgf.evaluate(*ep,cx,y);
  }
};

// The difference between two grid functions
template<typename Grid, typename A, typename B>
class MinusGridFunction
{
private:
  const A & a; // discrete grid function
  const B & b; // discrete grid function
  typedef typename A::Traits::DomainType DT;
  typedef typename A::Traits::RangeType RT;
  typedef typename Grid::template Codim<0>::Entity Entity;

public:
  
  typedef typename A::Traits Traits;
  
  MinusGridFunction(const A & a_, const B & b_)
    : a(a_), b(b_)
  {}

  void evaluate(const Entity & e, const DT & lx, RT & y) const
  {
    RT av; a.evaluate(e,lx,av);
    RT bv; b.evaluate(e,lx,bv);
    y = av - bv;
  }
};

// The product of two grid functions
template<typename Grid, typename A, typename B>
class ProductGridFunction
{
private:
  const A & a; // discrete grid function
  const B & b; // discrete grid function
  typedef typename A::Traits::DomainType DT;
  typedef typename A::Traits::RangeType RT;
  typedef typename Grid::template Codim<0>::Entity Entity;

public:
  
  typedef typename A::Traits Traits;
  
  ProductGridFunction(const A & a_, const B & b_)
    : a(a_), b(b_)
  {}

  void evaluate(const Entity & e, const DT & lx, RT & y) const
  {
    RT av; a.evaluate(e,lx,av);
    RT bv; b.evaluate(e,lx,bv);
    y = av * bv;
  }
};

/**
   This function may be used by the VTKWriter object to evaluate the
   functor object at a given local coordinate.
 */
template<class Grid, class Function>
class VTKFunctionWrapper : public Dune::VTKFunction<Grid>
{
private:

  // Some types which are exported by the base class
  typedef Dune::VTKFunction<Grid> Base;
  typedef typename Base::Entity Entity;
  typedef typename Base::ctype ctype;
  enum { dim =  Base::dim };

  // Constant reference to the functor object.
  const Function & f;
  const std::string data_name;
public:

  VTKFunctionWrapper(const Function & f_, std::string name_) : f(f_), data_name(name_)
  {}

  //! The number of components of this function. This is necessary as
  //! the VTKWriter would also allow the visualization of vector
  //! valued functions.
  int ncomps() const {return 1;}

  //! Evaluate the functor at a given local coordinate
  double evaluate(int comp, const Entity & e, const Dune::FieldVector<ctype,dim> & xi) const
  {
    typedef typename Entity::Geometry::GlobalCoordinate GlobalCoordinate;
    const GlobalCoordinate c = e.geometry().global(xi);
    typedef typename Function::Traits::RangeType RangeType;
    RangeType y;
    f.evaluateGlobal(c,y);
    return y[comp];
  }

  //! This name will later be displayed as a label for the visualized
  //! data in paraview.
  std::string name() const
  {
    return data_name;
  }
};


#endif // __DUNE_GRID_HOWTO_FUNCTORS_HH__
