// $Id: integration.cc 270 2010-01-14 14:47:44Z dedner $

// Dune includes
#include"config.h"           // file constructed by ./configure script
#include<dune/grid/sgrid.hh> // load sgrid definition
#include <dune/common/mpihelper.hh> // include mpi helper class 
#include <dune/common/configparser.hh> // include mpi helper class 
#include<dune/grid/onedgrid.hh>
#include<dune/grid/yaspgrid.hh>
#include <dune/grid/io/file/vtk/subsamplingvtkwriter.hh>
#if HAVE_UG 
#include<dune/grid/uggrid.hh>
#endif
#include<dune/grid/io/file/gmshreader.hh>
#include<sstream>

#include <dune/pdelab/common/function.hh>
#include <dune/pdelab/common/vtkexport.hh>
#include <dune/pdelab/backend/istlvectorbackend.hh>
#include <dune/pdelab/backend/istlmatrixbackend.hh>
#include <dune/pdelab/gridfunctionspace/gridfunctionspace.hh>
#include <dune/pdelab/gridfunctionspace/gridfunctionspaceutilities.hh>
#include <dune/pdelab/gridfunctionspace/interpolate.hh>
#include <dune/pdelab/gridfunctionspace/constraints.hh>
#include<dune/pdelab/finiteelementmap/pk2dfem.hh>
#include<dune/pdelab/finiteelementmap/pk3dfem.hh>
#include <dune/npde/finiteelement/pk1dbasis.hh>
#include <dune/npde/finiteelement/p12dbasis.hh>

#include <dune/localfunctions/monom.hh>
#include <dune/localfunctions/lagrange/pk.hh>

#include"functors.hh"
#include"integrateentity.hh"

template<typename Grid, int dim>
Grid* createSingleSimplexGrid()
{
  Dune::GridFactory<Grid> grid_factory;
  typedef typename Grid::template Codim<0>::Geometry::LocalCoordinate Coordinate;
  Coordinate vertex;

  vertex = 0;
  grid_factory.insertVertex(vertex);

  for(int i=0; i<dim; ++i){
    vertex = 0;
    vertex[i] = 1.0;
    grid_factory.insertVertex(vertex);
  }
  
  const int n_vertices = dim + 1;
  std::vector<unsigned int> indices(n_vertices);
  for(int i=0; i<n_vertices; ++i) 
    indices[i]=i;

  Dune::GeometryType simplex_type(Dune::GeometryType::simplex,dim);
  grid_factory.insertElement(simplex_type,indices);

  return grid_factory.createGrid();
}

//! A function demonstrating a finite element space on a grid view
template<class GV, class FEM>
void plot_basis (const GV& gv, const FEM & fem, int order)
{
  // coordinate and result type
  typedef typename GV::Grid::ctype Coord;
  typedef double Real;

  // make grid function space 
  typedef Dune::PDELab::NoConstraints CON;
  typedef Dune::PDELab::ISTLVectorBackend<1> VBE;
  typedef Dune::PDELab::GridFunctionSpace<GV,FEM,CON,VBE> GFS;
  GFS gfs(gv,fem);

  // make a std::vector of degree of freedom vectors
  typedef typename GFS::template VectorContainer<Real>::Type U;
  std::vector<U> u(gfs.globalSize(),U(gfs,0.0));

  // write a file for each global finite element function
  typedef Dune::PDELab::DiscreteGridFunction<GFS,U> DGF;
  std::vector<DGF *> udgf(gfs.globalSize());
  Dune::SubsamplingVTKWriter<GV> vtkwriter(gv,6);
  std::cout << "Writing VTK file for " << gfs.globalSize() << " basis functions" << std::endl;
  for (std::size_t i=0; i<u.size(); i++)
    {
      u[i][i] = 1.0;
      std::stringstream componentName;
      componentName << "comp_" << i;
      udgf[i] = new DGF(gfs,u[i]);
      vtkwriter.addVertexData(new Dune::PDELab::VTKGridFunctionAdapter<DGF>(*(udgf[i]),componentName.str()));
    }
  std::stringstream ss; ss << GV::dimension << "d_"<<  order;
  std::string filename("basis_"); filename += ss.str();
  vtkwriter.write("u",Dune::VTKOptions::binaryappended);
  for (std::size_t i=0; i<udgf.size(); i++) delete udgf[i];
}

//! uniform refinement test
template<class Grid, class GridFunction>
double uniformintegration (Grid& grid, const GridFunction & f, const int qorder, const int min_level,
                           const int max_level, const double relative_accuracy)
{
  // get GridView on leaf grid - type
  typedef typename Grid :: LevelGridView GridView;

  // loop over grid sequence
  double oldvalue=0;
  const std::string filename_base = "functor_";
  for (int k=min_level; k<max_level; k++)
    {
      // get GridView instance
      GridView gridView = grid.levelView(k);

      // Write VTK file
      std::ostringstream ss; ss << k;

      // compute integral with some order
      const double value = integrategrid(gridView,f,qorder);

      const double diff = std::abs(value-oldvalue);
      // print result and error estimate
      std::cout << "elements=" 
                << std::setw(8) << std::right
                << gridView.size(0)
                << " integral=" 
                << std::scientific << std::setprecision(12) 
                << value 
                << " difference=" << diff
                << std::endl;

      if(diff < relative_accuracy * value)
        return value;

      // save value of integral
      oldvalue=value;

    }

  std::cerr << "Could not achieve sufficient quadrature accuracy." << std::endl;
  exit(1);
  return 0;
}

//! A function demonstrating a finite element space on a grid view
template<class Grid, class FEM, class Function>
double interpolate_function (const Grid& grid, const int level, const FEM & fem, const Function & f, 
                             const int k, Dune::ConfigParser & configuration)
{
  // coordinate and result type
  typedef typename Grid::LevelGridView GV;
  typedef typename GV::Grid::ctype Coord;
  typedef double Real;

  // make grid function space 
  typedef Dune::PDELab::NoConstraints CON;
  typedef Dune::PDELab::ISTLVectorBackend<1> VBE;
  typedef Dune::PDELab::GridFunctionSpace<GV,FEM,CON,VBE> GFS;
  GV gv = grid.levelView(level);
  GFS gfs(gv,fem);
  std::cout << "Degrees of freedom: " << gfs.globalSize() << std::endl;

  // make a degree of freedom vector
  typedef typename GFS::template VectorContainer<Real>::Type U;
  U u(gfs,0.0);

  // interpolate from a given grid function
  Dune::PDELab::interpolate(f,gfs,u);

  // create discrete grid function 
  typedef Dune::PDELab::DiscreteGridFunction<GFS,U> DGF;
  DGF interpolated(gfs,u);

  { // Plot interpolation to VTK file
    std::stringstream ss; ss << Grid::dimension << "d_p"<< k << "_" << level;
    std::string filename("interpolated_");
    filename += ss.str();
    Dune::SubsamplingVTKWriter<GV> vtkwriter(gv,6);
    vtkwriter.addVertexData(new Dune::PDELab::VTKGridFunctionAdapter<DGF>(interpolated,"u"));
    vtkwriter.write(filename,Dune::VTKOptions::binaryappended);
  }

  // integration parameters
  const int quad_order = configuration.get<int>("integration.order");
  const double max_level = configuration.get<double>("integration.max_level");
  const double relative_accuracy = configuration.get<double>("integration.relative_accuracy");

  // Wrapper which allow to evaluate this function on a finer grid
  // level
  typedef GridLevelFunction<Grid,DGF> DLGF;
  DLGF wrap_interpolated(interpolated,level);

  typedef MinusGridFunction<Grid,DLGF,Function> L2DiffGridFunction;
  L2DiffGridFunction l2diff_gf(wrap_interpolated,f);
  typedef ProductGridFunction<Grid,L2DiffGridFunction,L2DiffGridFunction> L22ErrorGridFunction;
  L22ErrorGridFunction diffsquare_gf(l2diff_gf,l2diff_gf);

  // perform integration on whole grid beginning on the given level
  const double value = uniformintegration(grid,diffsquare_gf,quad_order,level,max_level,relative_accuracy);
  double error = std::sqrt(value);
  std::cout << "Error : " << error << std::endl;

  return error;
}

template<class Grid, class ReferenceFunction>
void interpolateOnLevels2d(Grid & grid, const ReferenceFunction & f, 
                         Dune::ConfigParser & configuration)
{
  // order
  const int k = 1;

  std::cout << "Interpolate with polynomial order " << k << std::endl;
  typedef typename Grid::LevelGridView GV;
  typedef typename Grid::ctype DF;
  typedef double RF;

  // The P1 local basis functions and their coefficients make the
  // finite element map
  typedef P12dLocalFiniteElementMap<DF,RF> FEM;

  // Plot basis functions in VTK files
  FEM fem_l0;
  plot_basis(grid.levelView(0),fem_l0,k);

  const int max_level = configuration.get<int>("interpolation.max_level");
  double old_error = 0.0;
  for(int level = 0; level <= max_level; ++level){

    const GV gv = grid.levelView(level);
    FEM fem;
    
    double error = interpolate_function(grid, level, fem, f, k, configuration);
    std::cout << "Convergence order (asymptotic): " << std::log(old_error / error)/std::log(2.0) << std::endl;
    old_error = error;
  }
}

int main(int argc, char **argv)
{
  // initialize MPI, finalize is done automatically on exit 
  Dune::MPIHelper::instance(argc,argv);

  Dune::ConfigParser configuration;
  configuration.parseFile("uebung07.ini");

  // start try/catch block to get error messages from dune
  try {
    using namespace Dune;
 
    { // 2d case

      // The dimension of the domain
      const int dim = 2;

      // Type  of the simplex grid
      typedef Dune::UGGrid<dim> Grid;

      // Create a grid consisting of a single simplex embedded in the
      // unit cube.
      Grid * pgrid = createSingleSimplexGrid<Grid,dim>();
      Grid & grid = *pgrid;

      typedef Grid::LevelGridView GV;
      typedef double RF;

      // Create grid levels.
      const double max_level = configuration.get<double>("integration.max_level");
      for(int l=0; l<max_level; ++l)
        grid.globalRefine(1);

      // Reference function 
      typedef Harmonic<GV,RF> ReferenceGridFunction;
      const int reference_host_level = 0;
      ReferenceGridFunction f(grid.levelView(reference_host_level));

      // A wrapper which allows to evaluate the reference function on
      // a finer grid level than the host level
      typedef GridLevelFunction<Grid,ReferenceGridFunction> ReferenceGridLevelFunction;
      ReferenceGridLevelFunction lf(f,reference_host_level);

      {
        // write vtk file
        Dune::SubsamplingVTKWriter<GV> vtkwriter(grid.levelView(reference_host_level),6);
        vtkwriter.addVertexData(new Dune::PDELab::VTKGridFunctionAdapter<ReferenceGridLevelFunction>(lf,"u"));
        vtkwriter.write("reference_2d",Dune::VTKOptions::binaryappended);
      }

      interpolateOnLevels2d<Grid,ReferenceGridFunction>(grid, f, configuration);

      delete pgrid;
    }

   
  }
  catch (std::exception & e) {
    std::cout << "STL ERROR: " << e.what() << std::endl;
    return 1;
  }
  catch (Dune::Exception & e) {
    std::cout << "DUNE ERROR: " << e.what() << std::endl;
    return 1;
  }
  catch (...) {
    std::cout << "Unknown ERROR" << std::endl;
    return 1;
  }

  // done
  return 0;
}
