#include <iostream>
#include "config.h"           // file constructed by ./configure script
#include <dune/common/mpihelper.hh> // include mpi helper class 
#include <dune/common/configparser.hh>  
#include <dune/grid/sgrid.hh> // load sgrid definition
#include <dune/grid/onedgrid.hh>
#include <dune/grid/io/file/vtk/subsamplingvtkwriter.hh>
#if HAVE_UG 
#include <dune/grid/uggrid.hh>
#endif
#include<dune/grid/io/file/gmshreader.hh>
#include<dune/istl/bvector.hh>
#include<dune/istl/operators.hh>
#include<dune/istl/solvers.hh>
#include<dune/istl/preconditioners.hh>
#include<dune/istl/io.hh>
#include<dune/istl/paamg/amg.hh>
#include <dune/pdelab/common/function.hh>
#include <dune/pdelab/common/functionutilities.hh>
#include <dune/pdelab/common/vtkexport.hh>
#include <dune/pdelab/backend/istlvectorbackend.hh>
#include <dune/pdelab/backend/istlmatrixbackend.hh>
#include <dune/pdelab/backend/istlsolverbackend.hh>
#include <dune/pdelab/gridfunctionspace/gridfunctionspace.hh>
#include <dune/pdelab/gridfunctionspace/gridfunctionspaceutilities.hh>
#include <dune/pdelab/gridfunctionspace/interpolate.hh>
#include <dune/pdelab/gridfunctionspace/constraints.hh>
#include <dune/pdelab/gridoperatorspace/gridoperatorspace.hh>
#include <dune/pdelab/stationary/linearproblem.hh>
#include<dune/pdelab/finiteelementmap/pk2dfem.hh>
#include<dune/pdelab/finiteelementmap/conformingconstraints.hh>

#include "../../dune/npde/finiteelement/pk1dbasis.hh"
#include "../../dune/npde/finiteelement/pk1dconstraints.hh"
#include "cdefem.hh"
#include "../../dune/npde/utilities/functionutilities.hh"


template<typename GV, typename RF>
class SharpEdgeProblem
{
  typedef ConvectionDiffusionBoundaryConditions::Type BCType;

  const double diffusion;

public:
  typedef ConvectionDiffusionParameterTraits<GV,RF> Traits;

  SharpEdgeProblem(const double diffusion_): diffusion(diffusion_)
  {}

  //! tensor diffusion coefficient
  typename Traits::PermTensorType
  A (const typename Traits::ElementType& e, const typename Traits::DomainType& x) const
  {
    typename Traits::PermTensorType I;
    for (std::size_t i=0; i<Traits::dimDomain; i++)
      for (std::size_t j=0; j<Traits::dimDomain; j++)
        I[i][j] = (i==j) ? diffusion : 0;
    return I;
  }

  //! velocity field
  typename Traits::RangeType
  b (const typename Traits::ElementType& e, const typename Traits::DomainType& x) const
  {
    typename Traits::RangeType v(1.0);
    return v;
  }

  //! sink term
  typename Traits::RangeFieldType 
  c (const typename Traits::ElementType& e, const typename Traits::DomainType& x) const
  {
    return 0.0;
  }

  //! source term
  typename Traits::RangeFieldType 
  f (const typename Traits::ElementType& e, const typename Traits::DomainType& x) const
  {
    return 0.0;
  }

  //! boundary condition type function
  BCType
  bctype (const typename Traits::IntersectionType& is, const typename Traits::IntersectionDomainType& x) const
  {
    typename Traits::DomainType xglobal = is.geometry().global(x);
    if(xglobal[0] < 1e-6  || xglobal[1] < 1e-6)
      return ConvectionDiffusionBoundaryConditions::Dirichlet;
    else
      return ConvectionDiffusionBoundaryConditions::Outflow;
  }

  //! Dirichlet boundary condition value
  typename Traits::RangeFieldType 
  g (const typename Traits::ElementType& e, const typename Traits::DomainType& xlocal) const
  {
    typename Traits::DomainType xglobal = e.geometry().global(xlocal);

    const double y = xglobal[1] - 0.3 - xglobal[0];

    if(y >= 0. &&  y <= 0.3)
      return 1.0;
    else 
      return 0.0;
  }

  //! Neumann boundary condition
  typename Traits::RangeFieldType 
  j (const typename Traits::IntersectionType& is, const typename Traits::IntersectionDomainType& x) const
  {
    return 0.0;
  }

  //! Neumann outflow boundary condition
  typename Traits::RangeFieldType 
  o (const typename Traits::IntersectionType& is, const typename Traits::IntersectionDomainType& x) const
  {
    return 0.0;
  }
};


//! solve the diffusion problem on a given GridView with a given finite element space
template<class GV, class P, class FEM, class CON>
void driver (const GV& gv, P & problem, FEM& fem, std::string filename_base, int& N, double& l2e)
{
  // coordinate and result type
  typedef typename GV::Grid::ctype Coord;
  typedef double Real;
  typedef P Problem;

  // make grid function space 
  typedef Dune::PDELab::ISTLVectorBackend<1> VBE;
  typedef Dune::PDELab::GridFunctionSpace<GV,FEM,CON,VBE> GFS;
  GFS gfs(gv,fem);
  N = gfs.globalSize();

  // make a std::vector of degree of freedom vectors and initialize it with Dirichlet extension
  typedef typename GFS::template VectorContainer<Real>::Type U;
  U u(gfs,0.0);
  typedef ConvectionDiffusionDirichletExtensionAdapter<Problem> G;
  G g(gv,problem);
  Dune::PDELab::interpolate(g,gfs,u);

  // make constraints container and initialize it
  typedef typename GFS::template ConstraintsContainer<Real>::Type CC;
  CC cc;
  ConvectionDiffusionBoundaryConditionAdapter<Problem> bctype(gv,problem);
  Dune::PDELab::constraints(bctype,gfs,cc);
  Dune::PDELab::set_nonconstrained_dofs(cc,0.0,u);
  std::cout << "constrained dofs=" << cc.size() << " of " << gfs.globalSize() << std::endl;

  // make local operator
  typedef ConvectionDiffusionFEM<Problem> LOP;
  LOP lop(problem);
  typedef Dune::PDELab::ISTLBCRSMatrixBackend<1,1> MBE;
  typedef Dune::PDELab::GridOperatorSpace<GFS,GFS,LOP,CC,CC,MBE> GOS;
  GOS gos(gfs,cc,gfs,cc,lop);

  // make linear solver and solve problem
  //typedef Dune::PDELab::ISTLBackend_SEQ_BCGS_SSOR LS;
  typedef Dune::PDELab::ISTLBackend_SEQ_SuperLU LS;
  LS ls(false);
  typedef Dune::PDELab::StationaryLinearProblemSolver<GOS,LS,U> SLP;
  SLP slp(gos,u,ls,1e-6);
  slp.apply();

  // compute errors
  typedef Dune::PDELab::DiscreteGridFunction<GFS,U> DGF;
  DGF udgf(gfs,u);
  typedef DifferenceSquaredAdapter<G,DGF> DifferenceSquared;
  DifferenceSquared differencesquared(g,udgf);
  typename DifferenceSquared::Traits::RangeType l2errorsquared(0.0);
  Dune::PDELab::integrateGridFunction(differencesquared,l2errorsquared,10);
  l2e = sqrt(l2errorsquared);

  // write vtk file
  typedef DifferenceAdapter<G,DGF> Difference;
  Difference difference(g,udgf);
  Dune::SubsamplingVTKWriter<GV> vtkwriter(gv,2);
  //Dune::VTKWriter<GV> vtkwriter(gv);
  vtkwriter.addVertexData(new Dune::PDELab::VTKGridFunctionAdapter<DGF>(udgf,"u_h"));
  vtkwriter.addVertexData(new Dune::PDELab::VTKGridFunctionAdapter<G>(g,"u"));
  vtkwriter.addVertexData(new Dune::PDELab::VTKGridFunctionAdapter<Difference>(difference,"u-u_h"));
  vtkwriter.write(filename_base,Dune::VTKOptions::binaryappended);
}

int main(int argc, char **argv)
{
  // initialize MPI, finalize is done automatically on exit 
  Dune::MPIHelper::instance(argc,argv);

  // start try/catch block to get error messages from dune
  try {

    Dune::ConfigParser configuration;
    configuration.parseFile("sharpedge.ini");

    // UGGrid/Pk2d version
#if HAVE_UG
    if (true)
    {
      // make uggrid
      typedef Dune::UGGrid<2> GridType;
      GridType grid;
      typedef std::vector<int> GmshIndexMap;
      GmshIndexMap boundary_index_map;
      GmshIndexMap element_index_map;
      std::string grid_file = configuration.get<std::string>("grid.filename");
      Dune::GmshReader<GridType> gmsh_reader;
      gmsh_reader.read(grid,grid_file,boundary_index_map,element_index_map,true,false);
      const int max_level = configuration.get<int>("grid.max_level");
      const int min_level = configuration.get<int>("grid.min_level");
      for(int i=0; i<max_level; ++i)grid.globalRefine(1);
      const int degree = configuration.get<int>("method.degree");
      
      std::vector<double> l2(max_level+1,0.0);
      std::vector<int> N(max_level+1,0);

      for (int i=min_level; i<=max_level; i++)
	{
	  typedef GridType::LevelGridView GV;
	  const GV& gv=grid.levelView(i);

          // make problem
          typedef SharpEdgeProblem<GV,double> PRO;
          PRO problem(configuration.get<double>("discretization.diffusion"));


	  // make finite element space and solve
	  typedef Dune::PDELab::ConformingDirichletConstraints CON;
	  if (degree==1) {
	    typedef Dune::PDELab::Pk2DLocalFiniteElementMap<GV,GridType::ctype,double,1> FEM;
	    FEM fem(gv);
	    std::stringstream filename;
	    filename << "ldomain_" << grid_file << "_l" << i << "_k" << degree;
	    driver<GV,PRO,FEM,CON>(gv,problem,fem,filename.str(),N[i],l2[i]);
	  }
	  if (degree==2) {
	    typedef Dune::PDELab::Pk2DLocalFiniteElementMap<GV,GridType::ctype,double,2> FEM;
	    FEM fem(gv);
	    std::stringstream filename;
	    filename << "ldomain_" << grid_file << "_l" << i << "_k" << degree;
	    driver<GV,PRO,FEM,CON>(gv,problem,fem,filename.str(),N[i],l2[i]);
	  }
	  if (degree==3) {
	    typedef Dune::PDELab::Pk2DLocalFiniteElementMap<GV,GridType::ctype,double,3> FEM;
	    FEM fem(gv);
	    std::stringstream filename;
	    filename << "ldomain_" << grid_file << "_l" << i << "_k" << degree;
	    driver<GV,PRO,FEM,CON>(gv,problem,fem,filename.str(),N[i],l2[i]);
	  }
	  if (degree==4) {
	    typedef Dune::PDELab::Pk2DLocalFiniteElementMap<GV,GridType::ctype,double,4> FEM;
	    FEM fem(gv);
	    std::stringstream filename;
	    filename << "ldomain_" << grid_file << "_l" << i << "_k" << degree;
	    driver<GV,PRO,FEM,CON>(gv,problem,fem,filename.str(),N[i],l2[i]);
	  }
	  if (degree==5) {
	    typedef Dune::PDELab::Pk2DLocalFiniteElementMap<GV,GridType::ctype,double,5> FEM;
	    FEM fem(gv);
	    std::stringstream filename;
	    filename << "ldomain_" << grid_file << "_l" << i << "_k" << degree;
	    driver<GV,PRO,FEM,CON>(gv,problem,fem,filename.str(),N[i],l2[i]);
	  }

	}

    std::cout << "Results on mesh=" << grid_file << " k=" << degree << std::endl; 
    for (int i=0; i<=max_level; i++) 
      {
	double rate1=0.0;
	if (i>0) rate1=log(l2[i]/l2[i-1])/log(0.5);
	std::cout << std::setw(12) << N[i]
		  << std::setw(12) << std::setprecision(4) << std::scientific << l2[i]
		  << std::setw(12) << std::setprecision(4) << std::scientific << rate1
		  << std::endl;
	}
    }
#endif
  }
  catch (std::exception & e) {
    std::cout << "STL ERROR: " << e.what() << std::endl;
    return 1;
  }
  catch (Dune::Exception & e) {
    std::cout << "DUNE ERROR: " << e.what() << std::endl;
    return 1;
  }
  catch (...) {
    std::cout << "Unknown ERROR" << std::endl;
    return 1;
  }

  // done
  return 0;
}
